package at.ac.tuwien.dbai.ges.solver;

import at.ac.tuwien.dbai.ges.schema.SchedulingHorizon;
import at.ac.tuwien.dbai.ges.solver.algorithm.AdaptiveChanceMap;
import at.ac.tuwien.dbai.ges.solver.algorithm.Algorithm;
import at.ac.tuwien.dbai.ges.solver.algorithm.SimulatedAnnealingTaskSolver;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.Move;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.RandomSelectionStrategy;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.SelectionStrategy;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.shift.*;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.task.AddOrRemoveTaskAssignment;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.task.ChangeTaskAssignment;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.task.SwapShiftAndTasksBetweenEmployees;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.task.SwapTaskAssignments;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.ForbiddenSequenceConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.ShiftSequenceConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.WorkloadConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.solver.OverlapConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.task.TaskRequirementConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.weight.HardConstraintWeightProvider;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionContext;
import at.ac.tuwien.dbai.ges.solver.converter.HorizonConverter;
import at.ac.tuwien.dbai.ges.solver.converter.ResultConverter;
import at.ac.tuwien.dbai.ges.solver.converter.WeightConverter;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;

import java.io.IOException;

/**
 * Reads an instances, sets up the simulated annealing algorithm and prints the result to a file.
 *
 * @author Lucas Kletzander
 */
public class Main {

    public static void main(String[] args) {

        if (args.length < 2) {
            System.out.println("Arguments: <problem-file> <solution-file-name> [<log-info-file-name>] [<log-values-file-name>]");
            return;
        }

        System.out.println("Parsing file...");

        try {

            // Could use seed as optional argument
            RandomProvider.init();

            // Read input
            SchedulingHorizon horizon = MarshallingUtil.readFromFile(args[0], SchedulingHorizon.class);

            ConversionContext context = new ConversionContext();
            Instance instance = new Instance();

            // Set up penalty weights for hard constraint violations
            HardConstraintWeightProvider provider = new HardConstraintWeightProvider(i -> 10.0 * i, i -> 10.0 * i);
            provider.intMapping.put(ShiftSequenceConstraint.class, i -> 100.0 * i);
            provider.intMapping.put(ForbiddenSequenceConstraint.class, i -> 100.0 * i);
            provider.intMapping.put(TaskRequirementConstraint.class, i -> 100.0 * i);
            provider.intMapping.put(OverlapConstraint.class, i -> 2.0 * instance.timeSlotLength * i);
            provider.intMapping.put(WorkloadConstraint.class, i -> .5 * i);

            context.weight = new WeightConverter(provider);

            // Convert input to internal format
            HorizonConverter converter = new HorizonConverter(context);
            converter.convert(horizon, instance);

            // Set log output
            LogPrinter info = new LogPrinter(args.length < 3 ? null : args[2], true);
            LogPrinter values = new LogPrinter(args.length < 4 ? null : args[3], true);

            // Set strategy for selecting moves
            SelectionStrategy strategy = new RandomSelectionStrategy(instance);

            // Set moves and their probabilities
            AdaptiveChanceMap<Move> moves = new AdaptiveChanceMap<>();
            moves.add(new AddOrRemoveShift(instance, strategy), 1);
            moves.add(new ChangeShiftLocal(instance, strategy), 1);
            moves.add(new ChangeShiftType(instance, strategy), 1);
            moves.add(new CreateSequence(instance, strategy), .01);
            moves.add(new SwapShiftsBetweenEmployees(instance, strategy), .01);
            moves.add(new SwapShiftsWithinEmployee(instance, strategy), .01);
            moves.add(new SwapPeriodBetweenEmployees(instance, strategy), .01);
            moves.add(new SwapPeriodWithinEmployee(instance, strategy), .01);
            moves.add(new AddOrRemoveTaskAssignment(instance, strategy), 1);
            moves.add(new ChangeTaskAssignment(instance, strategy), 1);
            moves.add(new SwapTaskAssignments(instance, strategy), 1);
            moves.add(new SwapShiftAndTasksBetweenEmployees(instance, strategy), .01);

            // Simulated annealing parameters
            double coolingRate = .995;
            double startTemp = 1000;

            // Start algorithm
            Algorithm algorithm = new SimulatedAnnealingTaskSolver(info, values, strategy, moves, coolingRate, startTemp);
            algorithm.apply(instance, context.solution);

            info.close();
            values.close();

            // Convert and print result
            ResultConverter resultConverter = new ResultConverter(context, args[0]);
            MarshallingUtil.printToFile(args[1], resultConverter.createAndConvert(context.solution));

            System.out.println("DONE");

        } catch (IOException e) {
            e.printStackTrace();
        }

    }

}
