package at.ac.tuwien.dbai.ges.solver.algorithm;

import at.ac.tuwien.dbai.ges.solver.RandomProvider;

import java.util.HashMap;
import java.util.Map;
import java.util.NavigableMap;
import java.util.TreeMap;

/**
 * Contains a map for selection of elements with specified probabilities.
 * Also allows to update these probabilities based on some measurement of hits and misses.
 * @param <T> The type of objects stored in the map.
 *
 * @author Lucas Kletzander
 */
public class AdaptiveChanceMap<T> {

    private NavigableMap<Double, T> map = new TreeMap<>();
    private Map<T, Double> reverseMap = new HashMap<>();
    private Map<T, Double> hitCount = new HashMap<>();
    private Map<T, Double> missCount = new HashMap<>();
    private boolean changed = false;

    public void add(T element, double value) {

        if (value > 0) {

            changed = true;
            reverseMap.put(element, value);
            hitCount.put(element, 1.0);
            missCount.put(element, 1.0);

        }

    }

    public void updateMap() {

        map.clear();
        double sum = reverseMap.values().stream().mapToDouble(i -> i).sum();

        double current = 0;
        for (Map.Entry<T, Double> entry : reverseMap.entrySet()) {
            current += entry.getValue() / sum;
            map.put(current, entry.getKey());
        }

        // to ensure numerical errors do not lead to problems
        map.put(1.0, map.lastEntry().getValue());
        changed = false;

    }

    public T getRandomElement() {

        if (changed) {
            updateMap();
        }

        return map.ceilingEntry(RandomProvider.getRandom().nextDouble()).getValue();

    }

    public void addHit(T element, double difference) {
        hitCount.compute(element, (k, v) -> v + difference);
    }

    public void addMiss(T element, double difference) {
        missCount.compute(element, (k, v) -> v + difference);
    }

    public void updateWithCount(boolean resetCount) {

        changed = true;

        for (T element : reverseMap.keySet()) {

            reverseMap.put(element, reverseMap.get(element) + hitCount.get(element) / missCount.get(element));

            if (resetCount) {
                hitCount.put(element, 1.0);
                missCount.put(element, 1.0);
            }

        }

    }

    public Map<Double, T> getChanceMap() {
        return map;
    }

}
