package at.ac.tuwien.dbai.ges.solver.algorithm;

import at.ac.tuwien.dbai.ges.solver.LogPrinter;
import at.ac.tuwien.dbai.ges.solver.RandomProvider;
import at.ac.tuwien.dbai.ges.solver.Util;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.Initialization;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.Move;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.BiasedSelectionStrategy2;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.SelectionStrategy;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.task.MinimizeShifts;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.*;

/**
 * This algorithm implements simulated annealing.
 *
 * @author Lucas Kletzander
 */
public class SimulatedAnnealingTaskSolver extends Algorithm {

    private final LogPrinter info;
    private final LogPrinter values;

    private final SelectionStrategy strategy;
    private final AdaptiveChanceMap<Move> moves;
    private final double coolingRate;
    private final double startTemp;

    public SimulatedAnnealingTaskSolver(LogPrinter info, LogPrinter values, SelectionStrategy strategy, AdaptiveChanceMap<Move> moves, double coolingRate, double startTemp) {

        this.info = info;
        this.values = values;
        this.strategy = strategy;
        this.moves = moves;
        this.coolingRate = coolingRate;
        this.startTemp = startTemp;

    }

    @Override
    public void apply(Instance instance, Solution solution) {

        info.println("Starting to work on instance with history " + instance.history + " and period length " + instance.periodLength);
        long startTime = System.currentTimeMillis();

        Algorithm minimizeShifts = new MinimizeShifts(instance);

        solution.value = new Initialization(instance).initialize(solution);

        instance.getConstraintStream(solution).forEach(i -> info.println(i.toString()));

        double t = startTemp;
        int lastChange = 0;

        int innerIterations = (int) instance.getConstraintStream(solution).count();
        info.println("Starting with " + innerIterations + " inner iterations");

        for (int i = 0; lastChange < 100; i++) {

            if (strategy instanceof BiasedSelectionStrategy2) {
                ((BiasedSelectionStrategy2) strategy).updateBias(solution);
            }

            for (int j = 0; j < instance.periodLength * solution.schedule.size() * i; j++) {

                Move move = moves.getRandomElement();
                if (!move.prepare(solution)) {
                    continue;
                }

                double change = move.evaluate();

                if (change <= 0 || RandomProvider.getRandom().nextDouble() < Math.exp(-change / t)) {

                    move.execute(solution);
                    solution.value += change;

                    if (change < 0) {
                        lastChange = Math.max(0, lastChange + (int) change);
                        //moves.addHit(move, 1);
                    }

                } else {
                    move.abort();
                    //moves.addMiss(move, 1);
                }

                if (j == 0 && instance.shiftDesign) {
                    minimizeShifts.apply(instance, solution);
                }

                if (solution.value == 0) {
                    break;
                }

            }

            if (solution.value == 0) {
                System.out.println("Stopping on 0-cost solution");
                break;
            }

            //moves.updateWithCount(true);

            lastChange++;
            t *= coolingRate;

            values.println(i + "," + lastChange + "," + t + "," + solution.value);

        }

        if (instance.shiftDesign) {
            minimizeShifts.apply(instance, solution);
        }

        info.println("RESULTS:");
        instance.getConstraintStream(solution).forEach(i -> info.println(i.toString()));

        // Update all shifts with their corresponding task assignments as during the algorithm these assignments are only updated via taskIndex.
        for (EmployeeSchedule schedule : solution.schedule.values()) {

            for (int i = instance.history; i < schedule.shifts.length; i++) {

                Shift shift = schedule.shifts[i];
                if (shift == null) {
                    continue;
                }

                int finalI = i;
                schedule.taskIndex.subMap(i * 24 * 60 + shift.start, i * 24 * 60 + shift.end).forEach((k, v) ->
                        v.forEach(j -> {
                                TaskInstance inst = solution.taskInstances.get(j);
                                inst.parts.forEach(l -> {
                                    Task task = new Task();
                                    task.demandId = l.demandId;
                                    task.start = Util.getTimeAtDay(inst.day, l.start, finalI);
                                    task.end = Util.getTimeAtDay(inst.day, l.end, finalI);
                                    shift.tasks.put(task.start, task);
                                });
                        }));

            }

        }

        info.println("Finished in " + (System.currentTimeMillis() - startTime));

        Move.count.forEach((k, v) -> info.println(k.getSimpleName() + ": " + v.executed + "/" + v.aborted));

        moves.getChanceMap().forEach((k, v) -> info.println(k + ": " + v.getClass().getSimpleName()));

    }

}
