package at.ac.tuwien.dbai.ges.solver.algorithm.move;

import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.*;
import at.ac.tuwien.dbai.ges.solver.definition.AbstractEmployee;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.Break;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Solution;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Stream;

/**
 * This is the base class for moves containing the methods prepare - evaluate - execute / abort.
 * It further contains several methods to access relevant constraints based on some filters.
 *
 * @author Lucas Kletzander
 */
public abstract class Move {

    public final Instance instance;
    public Set<Constraint> constraints = new HashSet<>();

    public static Map<Class, Count> count = new HashMap<>();

    public Move(Instance instance) {
        this.instance = instance;
    }

    /**
     * Prepares execution of the move on the solution.
     * @param solution The current solution.
     * @return <code>true</code> if the move was prepared successfully, <code>false</code> if the move can't be applied.
     */
    public abstract boolean prepare(Solution solution);

    /**
     * Evaluates the change in evaluation value that would be caused by the execution of this move.
     * @return The change in evaluation value.
     */
    public double evaluate() {
        return constraints.stream().mapToDouble(Constraint::evaluate).sum();
    }

    /**
     * Executes the current move.
     * @param solution The current solution.
     */
    public void execute(Solution solution) {
        count.computeIfAbsent(getClass(), k -> new Count()).executed++;
        constraints.forEach(Constraint::execute);
        constraints.clear();
    }

    /**
     * Aborts the current move and reverts back to the state before the move.
     */
    public void abort() {
        count.computeIfAbsent(getClass(), k -> new Count()).aborted++;
        constraints.forEach(Constraint::abort);
        constraints.clear();
    }

    public void processBreakConstraints(Break br, Consumer<BreakConstraint> function) {
        applyAndAdd(instance.getBreakConstraintStream(br), function);
    }

    public void processBreakArrangementConstraints(EmployeeSchedule schedule, int day, Consumer<BreakArrangementConstraint> function) {
        applyAndAdd(instance.getBreakArrangementConstraintStream(schedule, day), function);
    }

    public void processEmployeeConstraints(Consumer<EmployeeConstraint> function) {
        applyAndAdd(instance.getEmployeeConstraintStream(), function);
    }

    public void processShiftConstraints(String type, AbstractEmployee employee, int day, Consumer<ShiftConstraint> function) {
        applyAndAdd(instance.getShiftConstraintStream(type, employee, day), function);
    }

    public void processShiftArrangementConstraints(AbstractEmployee employee, Consumer<ShiftArrangementConstraint> function) {
        applyAndAdd(instance.getShiftArrangementConstraintStream(employee), function);
    }

    public void processTaskConstraints(String id, Set<String> skills, Consumer<TaskConstraint> function) {
        applyAndAdd(instance.getTaskConstraintStream(id, skills), function);
    }

    public void processShiftSwap(EmployeeSchedule fromSchedule, EmployeeSchedule toSchedule, int fromDay, int toDay) {

        if (fromSchedule.shifts[fromDay] != null) {

            processShiftConstraints(fromSchedule.shifts[fromDay].type, instance.employeeDefinition.getEmployee(fromSchedule.id), fromDay, i -> i.removeShift(fromSchedule.id, fromSchedule.shifts[fromDay]));
            processShiftConstraints(fromSchedule.shifts[fromDay].type, instance.employeeDefinition.getEmployee(toSchedule.id), toDay, i -> i.addShift(toSchedule.id, fromSchedule.shifts[fromDay]));

            fromSchedule.overlap.removeShift(fromDay, fromSchedule.shifts[fromDay]);
            toSchedule.overlap.addShift(toDay, fromSchedule.shifts[fromDay]);
            constraints.add(fromSchedule.overlap);
            constraints.add(toSchedule.overlap);

        }

    }

    private <T extends Constraint> void applyAndAdd(Stream<T> stream, Consumer<T> function) {

        stream.forEach(i -> {
            function.accept(i);
            constraints.add(i);
        });

    }

    public static void resetCount() {
        count.clear();
    }

    public class Count {
        public long executed;
        public long aborted;
    }

}
