package at.ac.tuwien.dbai.ges.solver.algorithm.move.selection;

import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.demand.TaskDemand;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Solution;
import at.ac.tuwien.dbai.ges.solver.solution.TaskInstance;

import java.util.HashMap;
import java.util.Map;
import java.util.stream.IntStream;

/**
 * A selection strategy that is biased towards areas with more conflicts (measured by higher values of certain constraints).
 * Uses cached values to prevent frequent constraint reevaluation.
 *
 * @author Lucas Kletzander
 */
public class BiasedSelectionStrategy2 implements SelectionStrategy {

    private final Instance instance;
    private final RandomSelectionStrategy randomStrategy;
    private double[] dayBias;
    private Map<String, Double> taskBias = new HashMap<>();

    public BiasedSelectionStrategy2(Instance instance) {
        this.instance = instance;
        randomStrategy = new RandomSelectionStrategy(instance);
        dayBias = new double[instance.history + instance.periodLength];
    }

    @Override
    public EmployeeSchedule selectEmployee(Solution solution) {
        return selectEmployee(solution, null);
    }

    @Override
    public EmployeeSchedule selectEmployeeExcept(Solution solution, EmployeeSchedule except) {
        return selectEmployee(solution, except);
    }

    private EmployeeSchedule selectEmployee(Solution solution, EmployeeSchedule except) {

        EmployeeSchedule schedule = randomStrategy.selectEmployeeExcept(solution, except);

        while (true) {
            EmployeeSchedule newSchedule = randomStrategy.selectEmployeeExcept(solution, except);
            if (newSchedule.overlap.value > schedule.overlap.value) {
                schedule = newSchedule;
            } else {
                break;
            }
        }

        return schedule;

    }

    @Override
    public int selectDay(Solution solution) {
        return selectDay(solution, -1);
    }

    @Override
    public int selectDayExcept(Solution solution, int except) {
        return selectDay(solution, except);
    }

    public int selectDay(Solution solution, int except) {

        int day = randomStrategy.selectDayExcept(solution, except);

        while (true) {
            int newDay = randomStrategy.selectDayExcept(solution, except);
            if (dayBias[newDay] > dayBias[day]) {
                day = newDay;
            } else {
                break;
            }
        }

        return day;

    }

    @Override
    public TaskInstance selectTask(Solution solution) {

        TaskInstance task = randomStrategy.selectTask(solution);

        while (true) {
            TaskInstance newTask = randomStrategy.selectTask(solution);
            if (taskBias.get(newTask.demandId) > taskBias.get(task.demandId)) {
                task = newTask;
            } else {
                break;
            }
        }

        return task;

    }

    public void updateBias(Solution solution) {

        IntStream.range(0, instance.periodLength).forEach(i -> dayBias[i] = instance.taskDemands.demand.get(i).stream().mapToDouble(TaskDemand::getConstraintValue).sum());
        solution.taskInstances.forEach((k, v) -> taskBias.put(k, instance.taskDemands.idMap.get(k).getConstraintValue()));

    }

}
