package at.ac.tuwien.dbai.ges.solver.algorithm.move.shift;

import at.ac.tuwien.dbai.ges.solver.RandomProvider;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.Move;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.SelectionStrategy;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;
import at.ac.tuwien.dbai.ges.solver.solution.Solution;

import java.util.HashMap;
import java.util.Map;

/**
 * This move swaps a period (max length 7) of shifts between different employees.
 *
 * @author Lucas Kletzander
 */
public class SwapPeriodBetweenEmployees extends Move {

    private final SelectionStrategy strategy;

    private int day1;
    private int day2;
    private EmployeeSchedule schedule1;
    private EmployeeSchedule schedule2;
    private Map<Integer, Shift> replacements1 = new HashMap<>();
    private Map<Integer, Shift> replacements2 = new HashMap<>();

    public SwapPeriodBetweenEmployees(Instance instance, SelectionStrategy strategy) {
        super(instance);
        this.strategy = strategy;
    }

    @Override
    public boolean prepare(Solution solution) {

        if (solution.schedule.size() < 2) {
            return false;
        }

        day1 = strategy.selectDay(solution);
        day2 = Math.min(Math.max(instance.history, day1 + RandomProvider.getRandom().nextInt(15) - 7), instance.getMaxDay() - 1);
        schedule1 = strategy.selectEmployee(solution);
        schedule2 = strategy.selectEmployeeExcept(solution, schedule1);

        replacements1.clear();
        replacements2.clear();

        for (int i = day1; i != day2; i += day1 < day2 ? 1 : -1) {

            if (schedule1.fixed[i] || schedule2.fixed[i]) {
                continue;
            }

            replacements1.put(i, schedule2.shifts[i]);
            replacements2.put(i, schedule1.shifts[i]);

            processShiftSwap(schedule1, schedule2, i, i);
            processShiftSwap(schedule2, schedule1, i, i);

        }

        processShiftArrangementConstraints(instance.employeeDefinition.getEmployee(schedule1.id), i -> i.changeRow(schedule1.id, schedule1.shifts, replacements1));
        processShiftArrangementConstraints(instance.employeeDefinition.getEmployee(schedule2.id), i -> i.changeRow(schedule2.id, schedule2.shifts, replacements2));

        return true;

    }

    @Override
    public void execute(Solution solution) {

        super.execute(solution);
        replacements1.forEach((k, v) -> schedule1.shifts[k] = v);
        replacements2.forEach((k, v) -> schedule2.shifts[k] = v);

    }

}
