package at.ac.tuwien.dbai.ges.solver.algorithm.move.task;

import at.ac.tuwien.dbai.ges.solver.algorithm.move.Move;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.SelectionStrategy;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;
import at.ac.tuwien.dbai.ges.solver.solution.Solution;
import at.ac.tuwien.dbai.ges.solver.solution.TaskInstance;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * This move swaps two shifts on the same day between employees and also swaps all tasks assigned at the time of the shifts.
 *
 * @author Lucas Kletzander
 */
public class SwapShiftAndTasksBetweenEmployees extends Move {

    private final SelectionStrategy strategy;

    private int day;
    private EmployeeSchedule schedule1;
    private EmployeeSchedule schedule2;
    private Map<Integer, Shift> replacements1 = new HashMap<>();
    private Map<Integer, Shift> replacements2 = new HashMap<>();
    private List<TaskMove> taskMoves = new LinkedList<>();

    public SwapShiftAndTasksBetweenEmployees(Instance instance, SelectionStrategy strategy) {
        super(instance);
        this.strategy = strategy;
    }

    @Override
    public boolean prepare(Solution solution) {

        if (solution.schedule.size() < 2) {
            return false;
        }

        day = strategy.selectDay(solution);
        schedule1 = strategy.selectEmployee(solution);
        schedule2 = strategy.selectEmployeeExcept(solution, schedule1);

        if (schedule1.fixed[day] || schedule2.fixed[day]) {
            return false;
        }

        replacements1.clear();
        replacements1.put(day, schedule2.shifts[day]);
        replacements2.clear();
        replacements2.put(day, schedule1.shifts[day]);

        processShiftSwap(schedule1, schedule2, day, day);
        processShiftSwap(schedule2, schedule1, day, day);

        processShiftArrangementConstraints(instance.employeeDefinition.getEmployee(schedule1.id), i -> i.changeRow(schedule1.id, schedule1.shifts, replacements1));
        processShiftArrangementConstraints(instance.employeeDefinition.getEmployee(schedule2.id), i -> i.changeRow(schedule2.id, schedule2.shifts, replacements2));

        taskMoves.clear();
        if (schedule1.shifts[day] != null) {
            schedule1.taskIndex.subMap(day * 24 * 60 + schedule1.shifts[day].start, day * 24 * 60 + schedule1.shifts[day].end).forEach((k, v) -> v.forEach(i -> moveTask(schedule1, schedule2, solution.taskInstances.get(i))));//TODO shifts overlapping days?
        }
        if (schedule2.shifts[day] != null) {
            schedule2.taskIndex.subMap(day * 24 * 60 + schedule2.shifts[day].start, day * 24 * 60 + schedule2.shifts[day].end).forEach((k, v) -> v.forEach(i -> moveTask(schedule2, schedule1, solution.taskInstances.get(i))));
        }

        constraints.add(schedule1.overlap);
        constraints.add(schedule2.overlap);

        return true;

    }

    private void moveTask(EmployeeSchedule fromSchedule, EmployeeSchedule toSchedule, TaskInstance task) {

        if (task.employees.contains(toSchedule.id)) {
            return;
        }

        processTaskConstraints(task.demandId, instance.employeeDefinition.getEmployee(fromSchedule.id).skills, i -> i.removeTask(task));
        processTaskConstraints(task.demandId, instance.employeeDefinition.getEmployee(toSchedule.id).skills, i -> i.addTask(task));

        task.parts.forEach(i -> {
            fromSchedule.overlap.removeOccupied(task.day, i.start, i.end);
            toSchedule.overlap.addOccupied(task.day, i.start, i.end);
        });

        taskMoves.add(new TaskMove(fromSchedule, toSchedule, task));

    }

    @Override
    public void execute(Solution solution) {

        super.execute(solution);

        schedule1.shifts[day] = replacements1.get(day);
        schedule2.shifts[day] = replacements2.get(day);

        taskMoves.forEach(i -> {
            i.task.employees.remove(i.fromSchedule.id);
            i.fromSchedule.getTasksAt(i.task).remove(i.task.demandId);
            i.task.employees.add(i.toSchedule.id);
            i.toSchedule.getTasksAt(i.task).add(i.task.demandId);
        });

    }

    private class TaskMove {

        EmployeeSchedule fromSchedule;
        EmployeeSchedule toSchedule;
        TaskInstance task;

        TaskMove(EmployeeSchedule fromSchedule, EmployeeSchedule toSchedule, TaskInstance task) {
            this.fromSchedule = fromSchedule;
            this.toSchedule = toSchedule;
            this.task = task;
        }

    }

}
