package at.ac.tuwien.dbai.ges.solver.algorithm.move.task;

import at.ac.tuwien.dbai.ges.solver.RandomProvider;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.Move;
import at.ac.tuwien.dbai.ges.solver.algorithm.move.selection.SelectionStrategy;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Solution;
import at.ac.tuwien.dbai.ges.solver.solution.TaskInstance;

import java.util.HashSet;
import java.util.Set;

/**
 * This move swaps a task assignment from one employee with overlapping task assignments from another employee.
 *
 * @author Lucas Kletzander
 */
public class SwapTaskAssignments extends Move {

    private final SelectionStrategy strategy;

    private TaskInstance task;
    private EmployeeSchedule schedule1;
    private EmployeeSchedule schedule2;

    public SwapTaskAssignments(Instance instance, SelectionStrategy strategy) {
        super(instance);
        this.strategy = strategy;
    }

    @Override
    public boolean prepare(Solution solution) {

        TaskInstance task = strategy.selectTask(solution);
        if (task.employees.isEmpty() || solution.schedule.isEmpty() || task.employees.size() == solution.schedule.size()) {
            return false;
        }

        EmployeeSchedule schedule1 = solution.schedule.get(RandomProvider.getRandomElement(task.employees));
        EmployeeSchedule schedule2;
        do {
            schedule2 = solution.scheduleList.get(RandomProvider.getRandom().nextInt(solution.scheduleList.size()));
        } while (schedule2.equals(schedule1) || task.employees.contains(schedule2.id));

        return prepare(solution, schedule1, schedule2, task.demandId);

    }

    public boolean prepare(Solution solution, EmployeeSchedule fromSchedule, EmployeeSchedule toSchedule, String taskId) {

        task = solution.taskInstances.get(taskId);
        schedule1 = fromSchedule;
        schedule2 = toSchedule;

        processTaskConstraints(task.demandId, instance.employeeDefinition.getEmployee(schedule1.id).skills, i -> i.removeTask(task));
        processTaskConstraints(task.demandId, instance.employeeDefinition.getEmployee(schedule2.id).skills, i -> i.addTask(task));

        task.parts.forEach(i -> {
            schedule1.overlap.removeOccupied(task.day, i.start, i.end);
            schedule2.overlap.addOccupied(task.day, i.start, i.end);
        });

        //TODO should adapt to improve usage for tasks that are split into several parts
        schedule2.taskIndex.subMap(task.day * 24 * 60 + task.getFirst().start, task.day * 24 * 60 + task.getLast().end).values().forEach(i -> i.forEach(j -> {

            TaskInstance t = solution.taskInstances.get(j);
            if (!t.employees.contains(schedule1.id)) {
                processTaskConstraints(t.demandId, instance.employeeDefinition.getEmployee(schedule2.id).skills, k -> k.removeTask(t));
                processTaskConstraints(t.demandId, instance.employeeDefinition.getEmployee(schedule1.id).skills, k -> k.addTask(t));

                t.parts.forEach(k -> {
                    schedule2.overlap.removeOccupied(t.day, k.start, k.end);
                    schedule1.overlap.addOccupied(t.day, k.start, k.end);
                });
            }

        }));

        constraints.add(schedule1.overlap);
        constraints.add(schedule2.overlap);

        return true;

    }

    @Override
    public void execute(Solution solution) {

        super.execute(solution);

        task.employees.remove(schedule1.id);
        schedule1.getTasksAt(task).remove(task.demandId);

        schedule2.taskIndex.subMap(task.day * 24 * 60 + task.getFirst().start, task.day * 24 * 60 + task.getLast().end).values().forEach(i -> {

            Set<String> keep = new HashSet<>();

            i.forEach(j -> {

                TaskInstance t = solution.taskInstances.get(j);
                if (!t.employees.contains(schedule1.id)) {
                    t.employees.remove(schedule2.id);
                    t.employees.add(schedule1.id);
                    schedule1.getTasksAt(t).add(t.demandId);
                } else {
                    keep.add(t.demandId);
                }

            });

            i.retainAll(keep);

        });

        task.employees.add(schedule2.id);
        schedule2.getTasksAt(task).add(task.demandId);

    }

}
