package at.ac.tuwien.dbai.ges.solver.constraint.hierarchy;

import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;

import java.util.Set;

/**
 * This is the base constraint class providing the methods to get the current value, evaluate the change and execute or abort a change.
 * The way to incorporate changes depends on the type of constraint.
 *
 * @author Lucas Kletzander
 */
public abstract class Constraint {

    public final Instance instance;

    public String label;
    public double value;
    public double newValue;

    public Set<String> types;
    public int startDay = -1;
    public int endDay = -1;
    public boolean[] validDays;

    public Constraint(Instance instance) {
        this.instance = instance;
    }

    /**
     * Returns the current constraint value.
     * @return The current value.
     */
    public double getValue() {
        return value;
    }

    /**
     * Evaluates the change in the constraint value.
     * @return The value change.
     */
    public double evaluate() {
        return newValue - value;
    }

    /**
     * Executes the change.
     */
    public void execute() {
        value = newValue;
    }

    /**
     * Aborts the change.
     */
    public void abort() {
        newValue = value;
    }

    @Override
    public String toString() {
        return (label == null ? getClass().getSimpleName() : label) + ": " + value;
    }

    public boolean applies(Shift shift) {
        return types == null || types.contains(shift.type);
    }

    public boolean applies(int day) {

        if (validDays == null) {
            return day >= startDay && (endDay == -1 || day <= endDay);
        } else {
            return validDays[day];
        }

    }

    public boolean applies(int start, int end) {
        return (start <= startDay && end >= startDay) || (start >= startDay && (endDay == -1 || start <= endDay));
    }

}
