package at.ac.tuwien.dbai.ges.solver.constraint.shift;

import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.ShiftArrangementConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.weight.WeightStrategy;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;

import java.util.*;

/**
 * Constrains the weekly minimum rest time.
 *
 * @author Lucas Kletzander
 */
public class MinWeekRestTimeConstraint extends ShiftArrangementConstraint {

    public int minDuration;
    public WeightStrategy<Integer> weightStrategy;
    public Map<String, NavigableMap<Integer, Integer>> rest = new HashMap<>();
    public Map<String, Map<Integer, Integer>> restChanges = new HashMap<>();

    public MinWeekRestTimeConstraint(Instance instance) {
        super(instance);
    }

    @Override
    public void changeRow(String id, Shift[] row, Map<Integer, Shift> replacements) {

        for (Integer i : replacements.keySet()) {

            int prev = getPrevious(row, replacements, i, Objects::nonNull);
            int next = getNext(row, i, Objects::nonNull);

            int endPrev = prev < 0 ? -minDuration : prev * 24 * 60 + replacements.getOrDefault(prev, row[prev]).end;
            int startNext = next < row.length ? next * 24 * 60 + row[next].start : instance.getMaxTime() + 2 * minDuration;

            if (row[i] == null) {
                removeFromRest(id, endPrev, startNext - endPrev, true, true);
            } else {
                removeFromRest(id, endPrev, i * 24 * 60 + row[i].start - endPrev, true, false);
                removeFromRest(id, i * 24 * 60 + row[i].end, startNext - i * 24 * 60 - row[i].end, false, true);
            }

            if (replacements.get(i) == null) {
                addToRest(id, endPrev, startNext - endPrev, true, true);
            } else {
                addToRest(id, endPrev, i * 24 * 60 + replacements.get(i).start - endPrev, true, false);
                addToRest(id, i * 24 * 60 + replacements.get(i).end, startNext - i * 24 * 60 - replacements.get(i).end, false, true);
            }

        }

    }

    @Override
    public void initRow(String id, Shift[] row, Map<Integer, Shift> replacements) {

        rest.computeIfAbsent(id, i -> new TreeMap<>()).put(-minDuration, instance.getMaxTime() + minDuration);
        super.initRow(id, row, replacements);

    }

    private void removeFromRest(String id, int start, int duration, boolean before, boolean after) {

        if (duration > minDuration) {
            restChanges.computeIfAbsent(id, i -> new HashMap<>()).putIfAbsent(start, rest.get(id).remove(start));
            if (before && start >= 0) {
                int startBefore = rest.get(id).lowerKey(start);
                newValue -= weightStrategy.evaluate(start - startBefore - rest.get(id).get(startBefore));
            }
            if (after && start + duration < instance.getMaxTime()) {
                newValue -= weightStrategy.evaluate(rest.get(id).higherKey(start) - start - duration);
            }
        }

    }

    private void addToRest(String id, int start, int duration, boolean before, boolean after) {

        if (duration > minDuration) {
            restChanges.computeIfAbsent(id, i -> new HashMap<>()).putIfAbsent(start, rest.get(id).put(start, duration));
            if (before && start >= 0) {
                int startBefore = rest.get(id).lowerKey(start);
                newValue += weightStrategy.evaluate(start - startBefore - rest.get(id).get(startBefore));
            }
            if (after && start + duration < instance.getMaxTime()) {
                newValue += weightStrategy.evaluate(rest.get(id).higherKey(start) - start - duration);
            }
        }

    }

    @Override
    public void execute() {
        super.execute();
        restChanges.clear();
    }

    @Override
    public void abort() {

        super.abort();

        restChanges.forEach((k, v) -> v.forEach((k1, v1) -> {
            if (v1 == null) {
                rest.get(k).remove(k1);
            } else {
                rest.get(k).put(k1, v1);
            }
        }));

        restChanges.clear();

    }

}
