package at.ac.tuwien.dbai.ges.solver.constraint.solver;

import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.Constraint;
import at.ac.tuwien.dbai.ges.solver.constraint.weight.WeightStrategy;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;

import java.util.stream.IntStream;

/**
 * Evaluates the overlap of shifts and tasks.
 *
 * @author Lucas Kletzander
 */
public class OverlapConstraint extends Constraint {

    public int[] availability;
    public int[] occupation;
    public int[] availabilityChange;
    public int[] availabilityOld;
    public int[] occupationChange;
    public int[] occupationOld;
    public int availabilityCount;
    public int occupationCount;
    public int count;
    public int newCount;
    public WeightStrategy<Integer> strategy;

    public OverlapConstraint(Instance instance) {

        super(instance);
        availability = new int[(instance.periodLength + 2) * 24 * 60 / instance.timeSlotLength];
        availabilityChange = new int[availability.length];
        availabilityOld = new int[availability.length];
        occupation = new int[(instance.periodLength + 2) * 24 * 60 / instance.timeSlotLength];
        occupationChange = new int[occupation.length];
        occupationOld = new int[occupation.length];

    }

    public void addShift(int day, Shift shift) {

        getSlots(day, shift.start, shift.end).forEach(i -> {

            availabilityChange[availabilityCount] = i;
            availabilityOld[availabilityCount++] = availability[i]++;

            if (occupation[i] >= availability[i]) {
                newCount--;
            }

            if (availability[i] > 1) {
                newCount++;
            }

        });

    }

    public void removeShift(int day, Shift shift) {

        getSlots(day, shift.start, shift.end).forEach(i -> {

            availabilityChange[availabilityCount] = i;
            availabilityOld[availabilityCount++] = availability[i]--;

            if (occupation[i] > availability[i]) {
                newCount++;
            }

            if (availability[i] == 1) {
                newCount--;
            }

        });

    }

    public void addOccupied(int day, int start, int end) {

        getSlots(day, start, end).forEach(i -> {

            occupationChange[occupationCount] = i;
            occupationOld[occupationCount++] = occupation[i]++;

            if (occupation[i] > availability[i]) {
                newCount++;
            }

        });

    }

    public void removeOccupied(int day, int start, int end) {

        getSlots(day, start, end).forEach(i -> {

            occupationChange[occupationCount] = i;
            occupationOld[occupationCount++] = occupation[i]--;

            if (occupation[i] >= availability[i]) {
                newCount--;
            }

        });

    }

    /**
     * Transforms a time period into the stream of slot indices where the planning period and one day on each end are included.
     * @param day The day in the extended planning period.
     * @param start The start time.
     * @param end The end time.
     * @return The stream of slot indices.
     */
    private IntStream getSlots(int day, int start, int end) {
        return day < instance.history - 1 ? IntStream.empty() : IntStream.range(((day - instance.history + 1) * 24 * 60 + start) / instance.timeSlotLength, ((day - instance.history + 1) * 24 * 60 + end) / instance.timeSlotLength);
    }

    @Override
    public double evaluate() {

        newValue = strategy.evaluate(newCount);
        return super.evaluate();

    }

    @Override
    public void execute() {

        super.execute();
        count = newCount;
        availabilityCount = 0;
        occupationCount = 0;

    }

    @Override
    public void abort() {

        super.abort();
        newCount = count;

        for (int i = availabilityCount - 1; i >= 0; i--) {
            availability[availabilityChange[i]] = availabilityOld[i];
        }
        availabilityCount = 0;

        for (int i = occupationCount - 1; i >= 0; i--) {
            occupation[occupationChange[i]] = occupationOld[i];
        }
        occupationCount = 0;

    }

}
