package at.ac.tuwien.dbai.ges.solver.converter;

import at.ac.tuwien.dbai.ges.solver.definition.Instance;

import java.lang.reflect.InvocationTargetException;

/**
 * A converter is used for converting from a source type to a target type.
 *
 * @author Lucas Kletzander
 */
public abstract class Converter<S,T> {

    public final Class<S> sourceClass;
    public final Class<T> targetClass;
    public final ConversionContext context;

    /**
     * Initializes the converter.
     * @param sourceClass The source class.
     * @param targetClass The target class.
     * @param context The conversion context.
     */
    public Converter(Class<S> sourceClass, Class<T> targetClass, ConversionContext context) {

        this.sourceClass = sourceClass;
        this.targetClass = targetClass;
        this.context = context;

    }

    /**
     * Has to be implemented by the converter to do the actual conversion.
     * @param source The source.
     * @param target The target.
     * @throws ConversionException If there is an error in the conversion.
     */
    public abstract void convert(S source, T target) throws ConversionException;

    /**
     * Creates a new object of the target class and performs the conversion.
     * @param source The source object, the check for the current type is done at runtime.
     * @return The newly created target, which is expected to have a constructor accepting precisely the instance.
     * @throws ConversionException If the source object has the wrong type, the target object can't be created or there is an error in the conversion.
     */
    @SuppressWarnings("unchecked")
    public T createAndConvert(Object source) throws ConversionException {

        try {

            T result;

            try {
                result = targetClass.newInstance();
            } catch (InstantiationException e) {
                result = targetClass.getConstructor(Instance.class).newInstance(context.instance);
            }

            if (!sourceClass.isAssignableFrom(source.getClass())) {
                throw new ConversionException("Source has wrong type " + source.getClass());
            }
            convert((S) source, result);

            return result;

        } catch (InstantiationException | IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
            throw new ConversionException("Error instantiating " + targetClass, e);
        }

    }

}
