package at.ac.tuwien.dbai.ges.solver.converter;

import at.ac.tuwien.dbai.ges.schema.*;
import at.ac.tuwien.dbai.ges.solver.converter.breaks.BreakConfigurationConverter;
import at.ac.tuwien.dbai.ges.solver.converter.breaks.BreakConverter;
import at.ac.tuwien.dbai.ges.solver.converter.breaks.BreakTypeConverter;
import at.ac.tuwien.dbai.ges.solver.converter.constraint.*;
import at.ac.tuwien.dbai.ges.solver.converter.demand.DemandsConverter;
import at.ac.tuwien.dbai.ges.solver.converter.demand.ShiftDemandConverter;
import at.ac.tuwien.dbai.ges.solver.converter.demand.TaskDemandConverter;
import at.ac.tuwien.dbai.ges.solver.converter.employee.ContractConverter;
import at.ac.tuwien.dbai.ges.solver.converter.employee.EmployeeConverter;
import at.ac.tuwien.dbai.ges.solver.converter.employee.SkillConverter;
import at.ac.tuwien.dbai.ges.solver.converter.shift.ShiftConverter;
import at.ac.tuwien.dbai.ges.solver.converter.task.TaskConverter;

import javax.xml.bind.JAXBElement;
import java.util.HashMap;
import java.util.Map;

/**
 * Manages the available converters.
 *
 * @author Lucas Kletzander
 */
public class ConverterProvider {

    public Map<Class, Converter> converterMap = new HashMap<>();
    public Map<String, Converter> jaxbConverterMap = new HashMap<>();

    /**
     * Initializes the provider by creating the default converters.
     * @param context The conversion context.
     */
    public ConverterProvider(ConversionContext context) {

        context.provider = this;
        converterMap.put(General.class, new GeneralConverter(context));
        converterMap.put(Tasks.class, new TaskConverter(context));

        converterMap.put(Shifts.class, new ShiftConverter(context));
        converterMap.put(ShiftTypes.ShiftType.EmployeeLimits.EmployeeLimit.class, new EmployeeLimitConverter(context));
        converterMap.put(GlobalShiftConstraints.AverageShiftLength.class, new AverageShiftLengthConverter(context));
        converterMap.put(GlobalShiftConstraints.ShiftInstances.class, new ShiftInstanceConverter(context));

        converterMap.put(Breaks.class, new BreakConverter(context));
        converterMap.put(BreakTypes.BreakType.class, new BreakTypeConverter(context));
        converterMap.put(BreakConfigurations.BreakConfiguration.class, new BreakConfigurationConverter(context));

        converterMap.put(Skills.class, new SkillConverter(context));
        converterMap.put(Contracts.class, new ContractConverter(context));

        jaxbConverterMap.put("countconstraint", new CountConstraintConverter(context));
        jaxbConverterMap.put("sequenceconstraint", new SequenceConstraintConverter(context));
        jaxbConverterMap.put("weekendcount", new WeekendCountConstraintConverter(context));
        jaxbConverterMap.put("weekendsequence", new WeekendSequenceConstraintConverter(context));
        jaxbConverterMap.put("forbiddensequence", new ForbiddenSequenceConstraintConverter(context));
        jaxbConverterMap.put("identicalsequence", new IdenticalSequenceConstraintConverter(context));

        converterMap.put(WeightedTimeSpan.class, new LargestWorkloadGapConstraintConverter(context));//TODO use more specific source type?

        jaxbConverterMap.put("minresttime", new MinRestTimeConverter(context));
        jaxbConverterMap.put("minweekresttime", new MinWeekRestTimeConverter(context));
        jaxbConverterMap.put("workload", new WorkloadConstraintConverter(context));
        jaxbConverterMap.put("shiftstarttimes", new ShiftStartTimesConstraintConverter(context));
        jaxbConverterMap.put("shiftendtimes", new ShiftEndTimesConstraintConverter(context));
        jaxbConverterMap.put("shiftlengths", new ShiftLengthsConstraintConverter(context));

        converterMap.put(Employees.class, new EmployeeConverter(context));
        converterMap.put(EmployeeList.VariableEmployee.class, new VariableEmployeeConstraintConverter(context));
        converterMap.put(AbstractEmployeeType.Preferences.ShiftOnRequest.class, new ShiftOnRequestConverter(context));
        converterMap.put(AbstractEmployeeType.Preferences.ShiftOffRequest.class, new ShiftOffRequestConverter(context));
        converterMap.put(PreAssignments.NoShift.class, new NoShiftConverter(context));
        jaxbConverterMap.put("pair", new ShiftPairConverter(context, true));
        jaxbConverterMap.put("notpair", new ShiftPairConverter(context, false));

        converterMap.put(Demands.class, new DemandsConverter(context));
        converterMap.put(ShiftCover.class, new ShiftDemandConverter(context));
        converterMap.put(TaskCover.class, new TaskDemandConverter(context));

    }

    /**
     * Returns a converter if available.
     * @param source The source class.
     * @param target The target class.
     * @param <S> The source type.
     * @param <T> The target type.
     * @return An applicable converter.
     * @throws ConversionException If no applicable converter is found.
     */
    public <S, T> Converter<S, T> getConverter(Class<S> source, Class<T> target) throws ConversionException {

        Converter c = converterMap.get(source);
        return checkConverter(c, source, target);

    }

    /**
     * Returns a converter if available.
     * @param source The source object, if of type <code>JAXBElement</code>, the converter is searched by element name, else by class.
     * @param target The target class.
     * @param <S> The source type.
     * @param <T> The target type.
     * @return An applicable converter.
     * @throws ConversionException If no applicable converter is found.
     */
    public <S, T> Converter<S, T> getConverter(S source, Class<T> target) throws ConversionException {

        if (source.getClass().equals(JAXBElement.class)) {
            return checkConverter(jaxbConverterMap.get(((JAXBElement) source).getName().getLocalPart().toLowerCase()), ((JAXBElement) source).getValue().getClass(), target);
        } else {
            return checkConverter(converterMap.get(source.getClass()), source.getClass(), target);
        }

    }

    /**
     * Checks if a converter is applicable.
     * @param converter The converter.
     * @param source The source class.
     * @param target The target class.
     * @param <S> The source type.
     * @param <T> The target type.
     * @return The converter if it is applicable.
     * @throws ConversionException If the converter is not applicable.
     */
    private <S, T> Converter<S, T> checkConverter(Converter converter, Class<?> source, Class<T> target) throws ConversionException {

        if (converter == null || !source.isAssignableFrom(converter.sourceClass) || !target.isAssignableFrom(converter.targetClass)) {
            throw new ConversionException("Could not find converter from " + source + " to " + target);
        }

        return converter;

    }

}
