package at.ac.tuwien.dbai.ges.solver.converter;

import at.ac.tuwien.dbai.ges.schema.DayCollection;
import at.ac.tuwien.dbai.ges.schema.DayRange;

import javax.xml.bind.JAXBElement;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.IntStream;

/**
 * Contains methods for converting various date and time notions.
 *
 * @author Lucas Kletzander
 */
public class DateTimeConverter {

    private static final Pattern PATTERN_TIME_POINT = Pattern.compile("(([0-9]{1,7})\\.)?([01]?[0-9]|2[0-3]):([0-5]?[0-9])");
    private static final Pattern PATTERN_TIME_SPAN = Pattern.compile("(([0-9]{1,7})\\.([01]?[0-9]|2[0-3])|([0-9]{1,8})):([0-5]?[0-9])|([0-9]{1,10})");

    public final int periodLength;
    public final int firstWeekDay;
    public int history;

    private final long startDate;
    private final DatatypeFactory datatypeFactory;

    /**
     * Initializes the converter with start and end date.
     * @param start The start date.
     * @param end The end date.
     * @throws ConversionException If there is an error initializing the datatype factory.
     */
    public DateTimeConverter(XMLGregorianCalendar start, XMLGregorianCalendar end) throws ConversionException {

        startDate = parseDate(start).getTime();
        periodLength = convertDate(end) + 1;
        firstWeekDay = (start.toGregorianCalendar().get(Calendar.DAY_OF_WEEK) + 7 - Calendar.MONDAY) % 7;

        try {
            datatypeFactory = DatatypeFactory.newInstance();
        } catch (DatatypeConfigurationException e) {
            throw new ConversionException("Error initializing DateTimeConverter", e);
        }

    }

    /**
     * Initializes the converter with the period length.
     * @param periodLength The length of the scheduling period.
     * @param firstWeekDay The weekday the scheduling period starts on (Monday = 0).
     * @throws ConversionException If there is an error initializing the datatype factory.
     */
    public DateTimeConverter(int periodLength, String firstWeekDay) throws ConversionException {

        startDate = -1;
        this.periodLength = periodLength;
        this.firstWeekDay = firstWeekDay == null ? 0 : convertWeekDay(firstWeekDay);

        try {
            datatypeFactory = DatatypeFactory.newInstance();
        } catch (DatatypeConfigurationException e) {
            throw new ConversionException("Error initializing DateTimeConverter", e);
        }

    }

    /**
     * Converts a date.
     * @param date The date in XML Gregorian calendar format.
     * @return The date as Java date.
     */
    public Date parseDate(XMLGregorianCalendar date) {
        return date.toGregorianCalendar().getTime();
    }

    /**
     * Converts a date to a day index.
     * @param date The date in XML Gregorian calendar format.
     * @return The day index of the date in the scheduling period.
     * @throws ConversionException If the scheduling period is not specified with start and end date.
     */
    public int convertDate(XMLGregorianCalendar date) throws ConversionException {

        if (startDate == -1) {
            throw new ConversionException("Error: Start date was not set");
        }
        return (int) ((parseDate(date).getTime() - startDate) / (1000 * 60 * 60 * 24));

    }

    /**
     * Converts a date or day specification to a day index.
     * @param dateOrDay The date of day.
     * @return The day index in the scheduling period.
     * @throws ConversionException If there is an error parsing the date or day.
     */
    public int convertDateOrDay(String dateOrDay) throws ConversionException {

        if (datatypeFactory == null) {
            throw new ConversionException("Error initializing DatatypeFactory");
        }

        try {
            if (dateOrDay.matches("-?\\d+")) {
                return Integer.parseInt(dateOrDay) + history;
            } else {
                return convertDate(datatypeFactory.newXMLGregorianCalendar(dateOrDay)) + history;
            }
        } catch (IllegalArgumentException e) {
            throw new ConversionException("Error parsing day " + dateOrDay, e);
        }

    }

    /**
     * Converts a time point into a time index.
     * @param timePoint The time point.
     * @return The time point in minutes since midnight.
     * @throws ConversionException If there is an error parsing the time point.
     */
    public static int convertTimePoint(String timePoint) throws ConversionException {

        Matcher matcher = PATTERN_TIME_POINT.matcher(timePoint);
        try {
            if (matcher.matches()) {
                return (matcher.group(2) != null ? Integer.parseInt(matcher.group(2)) * 24 * 60 : 0)
                        + Integer.parseInt(matcher.group(3)) * 60
                        + Integer.parseInt(matcher.group(4));
            } else {
                throw new ConversionException("Error parsing time point " + timePoint);
            }
        } catch (NumberFormatException e) {
            throw new ConversionException("Error parsing time point " + timePoint, e);
        }

    }

    /**
     * Converts a time span to minutes.
     * @param timeSpan The time span.
     * @return The time span in minutes.
     * @throws ConversionException If there is an error parsing the time span.
     */
    public static int convertTimeSpan(String timeSpan) throws ConversionException {

        Matcher matcher = PATTERN_TIME_SPAN.matcher(timeSpan);
        try {
            if (matcher.matches()) {
                return (matcher.group(1) != null ?
                        (matcher.group(2) != null ?
                                Integer.parseInt(matcher.group(2)) * 24 * 60 + Integer.parseInt(matcher.group(3)) * 60 :
                                Integer.parseInt(matcher.group(4)) * 60) +
                                Integer.parseInt(matcher.group(5)) :
                        Integer.parseInt(matcher.group(6)));
            } else {
                throw new ConversionException("Error parsing time span " + timeSpan);
            }
        } catch (NumberFormatException e) {
            throw new ConversionException("Error parsing time span " + timeSpan, e);
        }

    }

    /**
     * Converts a week day into the index of the weekday.
     * @param weekDay The weekday.
     * @return The index of the weekday where Monday is 0.
     * @throws ConversionException If there is an error parsing the weekday.
     */
    public static int convertWeekDay(String weekDay) throws ConversionException {

        try {
            return WeekDay.valueOf(weekDay.toUpperCase()).ordinal();
        } catch (IllegalArgumentException e) {
            throw new ConversionException(e);
        }

    }

    /**
     * Converts a day collection into a boolean array.
     * @param collection The day collection, if <code>null</code>, the result contains all days in the scheduling horizon.
     * @return A boolean array containing <code>true</code> for each day specified in the day collection.
     * @throws ConversionException If there is an error parsing the collection.
     */
    public boolean[] convertDayCollection(DayCollection collection) throws ConversionException {

        boolean[] c = new boolean[history + periodLength + 1];

        if (collection == null) {

            Arrays.fill(c, true);

        } else {

            buildCollection(c, collection.getDayOrWeekDayOrRange(), true);
            if (collection.getExclude() != null) {
                buildCollection(c, collection.getExclude().getDayOrWeekDayOrRange(), false);
            }

        }

        return c;

    }

    /**
     * Converts different options to specify days into a boolean array. At most one argument should be different from <code>null</code>.
     * @param dateOrDay A date or day or <code>null</code> to use another option.
     * @param weekDay A weekday or <code>null</code> to use another option.
     * @param dayCollection A day collection or <code>null</code> to use another option.
     * @return A boolean array containing <code>true</code> for each day specified by the options or for all days if no options are different from <code>null</code>.
     */
    public boolean[] convertDayOptions(String dateOrDay, String weekDay, DayCollection dayCollection) {

        boolean[] c;

        if (dateOrDay != null) {

            c = new boolean[history + periodLength + 1];
            c[convertDateOrDay(dateOrDay)] = true;

        } else if (weekDay != null) {

            c = new boolean[history + periodLength + 1];
            int day = (convertWeekDay(weekDay) + 7 - firstWeekDay) % 7 + history;
            while (day < c.length) {
                c[day] = true;
                day += 7;
            }

        } else {

            c = convertDayCollection(dayCollection);

        }

        return c;

    }

    /**
     * Converts different options used within a day collection.
     * @param collection The result array.
     * @param list The list of options.
     * @param value <code>true</code> to add, <code>false</code> to remove elements from the result.
     * @throws ConversionException If the collection contains unknown elements.
     */
    private void buildCollection(boolean[] collection, List<JAXBElement<?>> list, boolean value) throws ConversionException {

        for (JAXBElement<?> element : list) {

            switch (element.getName().getLocalPart().toLowerCase()) {
                case "day":
                    collection[convertDateOrDay(element.getValue().toString())] = value;
                    break;
                case "weekday":
                    int weekDay = (convertWeekDay(element.getValue().toString()) + 7 - firstWeekDay) % 7 + history;
                    while (weekDay < collection.length) {
                        collection[weekDay] = value;
                        weekDay += 7;
                    }
                    break;
                case "range":
                    DayRange range = (DayRange) element.getValue();
                    IntStream.rangeClosed(range.getStartDay() == null ? 0 : convertDateOrDay(range.getStartDay()),
                            range.getEndDay() == null ? periodLength - 1 : convertDateOrDay(range.getEndDay())).forEach(i -> collection[i] = value);
                    break;
                default:
                    throw new ConversionException("Unknown collection element " + element.getName());
            }

        }

    }

    /**
     * Specifies weekdays and their corresponding ordinals.
     */
    public enum WeekDay {
        MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY
    }

}
