package at.ac.tuwien.dbai.ges.solver.converter;

import at.ac.tuwien.dbai.ges.schema.General;
import at.ac.tuwien.dbai.ges.schema.PreAssignments;
import at.ac.tuwien.dbai.ges.solver.definition.Instance;

/**
 * Converts the general section.
 *
 * @author Lucas Kletzander
 */
public class GeneralConverter extends Converter<General, Instance> {

    public GeneralConverter(ConversionContext context) {
        super(General.class, Instance.class, context);
    }

    @Override
    public void convert(General source, Instance target) throws ConversionException {

        if ((source.getStartDate() == null || source.getEndDate() == null) && source.getPeriodLength() == null) {
            throw new ConversionException("Must specify either start and end date or period length");
        }

        if (source.getStartDate() == null) {
            context.dateTime = new DateTimeConverter(source.getPeriodLength().intValue(), source.getFirstWeekDay());
        } else {
            context.dateTime = new DateTimeConverter(source.getStartDate(), source.getEndDate());
        }

        target.history = Math.max(0, -context.horizon.getEmployees().getEmployeeList().getEmployeeOrVariableEmployee().stream()
                .filter(i -> i.getPreAssignments() != null)
                .flatMap(i -> i.getPreAssignments().getShiftOrNoShift().stream()
                        .filter(j -> j instanceof PreAssignments.Shift)
                        .map(j -> ((PreAssignments.Shift) j).getDay()))
                .mapToInt(i -> context.dateTime.convertDateOrDay(i)).min().orElse(0));
        context.dateTime.history = target.history;

        target.periodLength = context.dateTime.periodLength;
        target.firstWeekDay = context.dateTime.firstWeekDay;

        if (source.getTimeSlotLength() != null) {
            target.timeSlotLength = DateTimeConverter.convertTimeSpan(source.getTimeSlotLength());
        }
        if (target.shiftSlotLength < 1) {
            target.shiftSlotLength = target.timeSlotLength;
        }

        if (source.getStartTime() != null) {
            target.startTime = DateTimeConverter.convertTimePoint(source.getStartTime());
        }

        if (source.getEndTime() != null) {
            target.endTime = DateTimeConverter.convertTimePoint(source.getEndTime());
        }

        if (source.getWeekendDefinition() != null) {
            convertWeekend(source.getWeekendDefinition(), target.weekend);
        }

        if (source.isIgnoreIndividualRosters() != null) {
            target.ignoreIndividualRosters = source.isIgnoreIndividualRosters();
        }

        if (source.isShiftDesign() != null) {
            target.shiftDesign = source.isShiftDesign();
        }

        if (source.getCyclicSchedule() != null) {
            try {
                target.cyclicSchedule = Instance.CyclicSchedule.valueOf(source.getCyclicSchedule().toUpperCase());
            } catch (IllegalArgumentException e) {
                throw new ConversionException(e);
            }
        }

        General.AllowSequenceCutoff allowSequenceCutoff = source.getAllowSequenceCutoff();
        if (allowSequenceCutoff != null) {
            target.cutoffStart = allowSequenceCutoff.isHorizonStart();
            target.cutoffEnd = allowSequenceCutoff.isHorizonEnd();
        }
        
    }

    private void convertWeekend(General.WeekendDefinition weekendDefinition, Instance.Weekend weekend) throws ConversionException {

        weekend.isSimple = weekendDefinition.getPrecise() == null;

        if (weekend.isSimple) {
            weekend.isWeekend = new boolean[7];
            weekendDefinition.getDays().forEach(d -> weekend.isWeekend[DateTimeConverter.convertWeekDay(d)] = true);
            weekend.updateWeekendStartSimple();
        } else {
            weekend.startWeekDay = DateTimeConverter.convertWeekDay(weekendDefinition.getPrecise().getStartDay());
            weekend.startTime = DateTimeConverter.convertTimePoint(weekendDefinition.getPrecise().getStartTime());
            weekend.startCompleteOnly = weekendDefinition.getPrecise().isStartCompleteOnly();
            weekend.endWeekDay = DateTimeConverter.convertWeekDay(weekendDefinition.getPrecise().getEndDay());
            weekend.endTime = DateTimeConverter.convertTimePoint(weekendDefinition.getPrecise().getEndTime());
            weekend.endCompleteOnly = weekendDefinition.getPrecise().isEndCompleteOnly();
        }

    }
    
}
