package at.ac.tuwien.dbai.ges.solver.converter;

import at.ac.tuwien.dbai.ges.schema.*;
import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.Constraint;
import at.ac.tuwien.dbai.ges.solver.constraint.weight.HardConstraintWeightProvider;
import at.ac.tuwien.dbai.ges.solver.constraint.weight.WeightStrategy;

import javax.xml.bind.JAXBElement;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

/**
 * Converts various notions for constraint weights into weight strategies.
 *
 * @author Lucas Kletzander
 */
public class WeightConverter {

    public final HardConstraintWeightProvider hardConstraintWeight;

    /**
     * Initializes the weight converter.
     * @param hardConstraintWeight The provider for weights assigned to hard constraints.
     */
    public WeightConverter(HardConstraintWeightProvider hardConstraintWeight) {
        this.hardConstraintWeight = hardConstraintWeight;
    }

    /**
     * Converts a weighted time point into a weight strategy accepting integer values.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param timePoint The weighted time point.
     * @param isMin <code>true</code> if the time point is a minimum, <code>false</code> for a maximum.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Integer> convertWeightedTimePointInt(Constraint constraint, WeightedTimePoint timePoint, boolean isMin) {

        return new WeightStrategy<>(convertMinOrMaxInt(DateTimeConverter.convertTimePoint(timePoint.getValue()), isMin),
                timePoint.getWeight() == null ? hardConstraintWeight.getIntWeightFunction(constraint.getClass()) : convertWeight(timePoint.getFunction(), timePoint.getWeight()),
                isMin);

    }

    /**
     * Converts a weighted time span into a weight strategy accepting integer values.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param timeSpan The weighted time span.
     * @param isMin <code>true</code> if the time point is a minimum, <code>false</code> for a maximum.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Integer> convertWeightedTimeSpanInt(Constraint constraint, WeightedTimeSpan timeSpan, boolean isMin) {

        return new WeightStrategy<>(convertMinOrMaxInt(DateTimeConverter.convertTimeSpan(timeSpan.getValue()), isMin),
                timeSpan.getWeight() == null ? hardConstraintWeight.getIntWeightFunction(constraint.getClass()) : convertWeight(timeSpan.getFunction(), timeSpan.getWeight()),
                isMin);

    }

    /**
     * Converts a weighted time span into a weight strategy accepting double values.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param timeSpan The weighted time span.
     * @param isMin <code>true</code> if the time point is a minimum, <code>false</code> for a maximum.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Double> convertWeightedTimeSpanDouble(Constraint constraint, WeightedTimeSpan timeSpan, boolean isMin) {

        return new WeightStrategy<>(convertMinOrMaxDouble(DateTimeConverter.convertTimeSpan(timeSpan.getValue()), isMin),
                timeSpan.getWeight() == null ? hardConstraintWeight.getDoubleWeightFunction(constraint.getClass()) : convertWeight(timeSpan.getFunction(), timeSpan.getWeight()),
                isMin);

    }

    /**
     * Converts a hard time span into a weight strategy accepting integer values.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param timeSpan The time span.
     * @param isMin <code>true</code> if the time point is a minimum, <code>false</code> for a maximum.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Integer> convertHardTimeSpanInt(Constraint constraint, String timeSpan, boolean isMin) {

        return new WeightStrategy<>(convertMinOrMaxInt(DateTimeConverter.convertTimeSpan(timeSpan), isMin),
                hardConstraintWeight.getIntWeightFunction(constraint.getClass()),
                isMin);

    }

    /**
     * Converts a weighted integer into a weight strategy accepting integer values.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param integer The weighted integer.
     * @param isMin <code>true</code> if the time point is a minimum, <code>false</code> for a maximum.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Integer> convertWeightedInteger(Constraint constraint, WeightedInteger integer, boolean isMin) {

        return new WeightStrategy<>(convertMinOrMaxInt(integer.getValue().intValue(), isMin),
                integer.getWeight() == null ? hardConstraintWeight.getIntWeightFunction(constraint.getClass()) : convertWeight(integer.getFunction(), integer.getWeight()),
                isMin);

    }

    /**
     * Converts a weighted double into a weight strategy accepting double values.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param dbl The weighted double.
     * @param isMin <code>true</code> if the time point is a minimum, <code>false</code> for a maximum.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Double> convertWeightedDouble(Constraint constraint, WeightedDouble dbl, boolean isMin) {

        return new WeightStrategy<>(convertMinOrMaxDouble(dbl.getValue(), isMin),
                dbl.getWeight() == null ? hardConstraintWeight.getDoubleWeightFunction(constraint.getClass()) : convertWeight(dbl.getFunction(), dbl.getWeight()),
                isMin);

    }

    /**
     * Returns a weight strategy for a fixed hard constraint.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @return The corresponding weight strategy.
     */
    public WeightStrategy<Integer> convertHardInt(Constraint constraint) {
        return new WeightStrategy<>(i -> i, hardConstraintWeight.getIntWeightFunction(constraint.getClass()), false);
    }

    /**
     * Converts the given function type into the actual function used for weighting.
     * @param function The given function type.
     * @param weight The weight to use for the function.
     * @param <T> The type of Number to accept.
     * @return A function applying the specified weighting function.
     * @throws ConversionException If the function type is unknown.
     */
    public <T extends Number> Function<T, Double> convertWeight(String function, double weight) throws ConversionException {

        if (function == null) {
            function = "linear";
        }

        switch (function.toLowerCase()) {
            case "constant":
                return value -> value.doubleValue() > 0.0 ? weight : 0;
            case "linear":
                return value -> weight * value.doubleValue();
            case "quadratic":
                return value -> weight * value.doubleValue() * value.doubleValue();
            default:
                throw new ConversionException("Weighting function " + function + " not recognized");
        }

    }

    /**
     * Converts an integer threshold into a function returning the distance to the threshold.
     * @param threshold The threshold.
     * @param isMin <code>true</code> if threshold is a minimum, <code>false</code> for a maximum.
     * @return A function returning the distance to the threshold if threshold is violated.
     */
    public Function<Integer, Integer> convertMinOrMaxInt(int threshold, boolean isMin) {

        if (isMin) {
            return value -> Math.max(threshold - value, 0);
        } else {
            return value -> Math.max(value - threshold, 0);
        }

    }

    /**
     * Converts an integer threshold into a function returning the distance to the threshold.
     * @param threshold The threshold.
     * @param isMin <code>true</code> if threshold is a minimum, <code>false</code> for a maximum.
     * @return A function returning the distance to the threshold if threshold is violated.
     */
    public Function<Double, Double> convertMinOrMaxDouble(int threshold, boolean isMin) {

        if (isMin) {
            return value -> Math.max(threshold - value, 0);
        } else {
            return value -> Math.max(value - threshold, 0);
        }

    }

    /**
     * Converts a double threshold into a function returning the distance to the threshold.
     * @param threshold The threshold.
     * @param isMin <code>true</code> if threshold is a minimum, <code>false</code> for a maximum.
     * @return A function returning the distance to the threshold if threshold is violated.
     */
    public Function<Double, Double> convertMinOrMaxDouble(double threshold, boolean isMin) {

        if (isMin) {
            return value -> Math.max(threshold - value, 0);
        } else {
            return value -> Math.max(value - threshold, 0);
        }

    }

    /**
     * Converts a list of weighted integers describing minimum and maximum values into a list of weight strategies.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param list The list of weighted integers.
     * @return The list of weight strategies.
     */
    public List<WeightStrategy<Integer>> convertMinMaxStrategyList(Constraint constraint, List<JAXBElement<WeightedInteger>> list) {

        List<WeightStrategy<Integer>> result = new ArrayList<>();

        for (JAXBElement<WeightedInteger> threshold : list) {

            switch (threshold.getName().getLocalPart().toLowerCase()) {
                case "min":
                    result.add(convertWeightedInteger(constraint, threshold.getValue(), true));
                    break;
                case "max":
                    result.add(convertWeightedInteger(constraint, threshold.getValue(), false));
                    break;
                default:
                    throw new ConversionException("Unknown element " + threshold.getName());
            }

        }

        return result;

    }

    /**
     * Converts a list of weighted time points describing minimum and maximum values into a list of weight strategies.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param list The list of weighted time points.
     * @return The list of weight strategies.
     */
    public List<WeightStrategy<Integer>> convertMinMaxPointStrategyList(Constraint constraint, List<JAXBElement<WeightedTimePoint>> list) {

        List<WeightStrategy<Integer>> result = new ArrayList<>();

        for (JAXBElement<WeightedTimePoint> threshold : list) {

            switch (threshold.getName().getLocalPart().toLowerCase()) {
                case "min":
                    result.add(convertWeightedTimePointInt(constraint, threshold.getValue(), true));
                    break;
                case "max":
                    result.add(convertWeightedTimePointInt(constraint, threshold.getValue(), false));
                    break;
                default:
                    throw new ConversionException("Unknown element " + threshold.getName());
            }

        }

        return result;

    }

    /**
     * Converts a list of weighted time spans describing minimum and maximum values into a list of weight strategies.
     * @param constraint The constraint for retrieval of a hard constraint strategy.
     * @param list The list of weighted time spans.
     * @return The list of weight strategies.
     */
    public List<WeightStrategy<Integer>> convertMinMaxSpanStrategyList(Constraint constraint, List<JAXBElement<WeightedTimeSpan>> list) {

        List<WeightStrategy<Integer>> result = new ArrayList<>();

        for (JAXBElement<WeightedTimeSpan> threshold : list) {

            switch (threshold.getName().getLocalPart().toLowerCase()) {
                case "min":
                    result.add(convertWeightedTimeSpanInt(constraint, threshold.getValue(), true));
                    break;
                case "max":
                    result.add(convertWeightedTimeSpanInt(constraint, threshold.getValue(), false));
                    break;
                default:
                    throw new ConversionException("Unknown element " + threshold.getName());
            }

        }

        return result;

    }

}
