package at.ac.tuwien.dbai.ges.solver.converter.breaks;

import at.ac.tuwien.dbai.ges.schema.BreakConfigurations;
import at.ac.tuwien.dbai.ges.solver.constraint.breaks.BreakTimeFractionConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.breaks.TotalBreakTimeConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.breaks.WorkAfterLastConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.breaks.WorkBeforeFirstConstraint;
import at.ac.tuwien.dbai.ges.solver.converter.*;
import at.ac.tuwien.dbai.ges.solver.definition.BreakConfiguration;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

/**
 * Converts definitions of break configurations.
 *
 * @author Lucas Kletzander
 */
public class BreakConfigurationConverter extends Converter<BreakConfigurations.BreakConfiguration, BreakConfiguration> {

    public BreakConfigurationConverter(ConversionContext context) {
        super(BreakConfigurations.BreakConfiguration.class, BreakConfiguration.class, context);
    }

    @Override
    public void convert(BreakConfigurations.BreakConfiguration source, BreakConfiguration target) throws ConversionException {

        BreakConfigurations.BreakConfiguration.ShiftFilter f = source.getShiftFilter();

        if (f != null) {

            BreakConfiguration.ShiftFilter filter = new BreakConfiguration.ShiftFilter();
            filter.validDays = context.dateTime.convertDayOptions(f.getDay(), f.getWeekDay(), f.getDays());

            if (f.getMinShiftLength() != null) {
                filter.minShiftLength = DateTimeConverter.convertTimeSpan(f.getMinShiftLength());
            }
            if (f.getMaxShiftLength() != null) {
                filter.maxShiftLength = DateTimeConverter.convertTimeSpan(f.getMaxShiftLength());
            }
            if (f.getMinShiftStart() != null) {
                filter.minShiftStart = DateTimeConverter.convertTimePoint(f.getMinShiftStart());
            }
            if (f.getMaxShiftStart() != null) {
                filter.maxShiftStart = DateTimeConverter.convertTimePoint(f.getMaxShiftStart());
            }
            if (f.getMinShiftEnd() != null) {
                filter.minShiftEnd = DateTimeConverter.convertTimePoint(f.getMinShiftEnd());
            }
            if (f.getMaxShiftEnd() != null) {
                filter.maxShiftEnd = DateTimeConverter.convertTimePoint(f.getMaxShiftEnd());
            }
            if (f.getContracts() != null) {
                filter.contracts = new HashSet<>(f.getContracts().getContract());//TODO contract parents
            }
            if (f.getShiftTypes() != null) {
                filter.types = ConverterUtil.convertShiftTypes(context.instance.shiftDefinition, f.getShiftTypes());
            }
            if (f.getMinWorkLength() != null) {
                filter.minWorkLength = DateTimeConverter.convertTimeSpan(f.getMinWorkLength());
            }
            if (f.getMaxWorkLength() != null) {
                filter.maxWorkLength = DateTimeConverter.convertTimeSpan(f.getMaxWorkLength());
            }

            target.filter = filter;

        }

        target.constraints = new ArrayList<>();

        if (source.getMinWorkBeforeFirst() != null) {
            WorkBeforeFirstConstraint c = new WorkBeforeFirstConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedTimeSpanInt(c, source.getMinWorkBeforeFirst(), true);
            target.constraints.add(c);
        }

        if (source.getMaxWorkBeforeFirst() != null) {
            WorkBeforeFirstConstraint c = new WorkBeforeFirstConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedTimeSpanInt(c, source.getMaxWorkBeforeFirst(), false);
            target.constraints.add(c);
        }

        if (source.getMinWorkAfterLast() != null) {
            WorkAfterLastConstraint c = new WorkAfterLastConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedTimeSpanInt(c, source.getMinWorkAfterLast(), true);
            target.constraints.add(c);
        }

        if (source.getMaxWorkAfterLast() != null) {
            WorkAfterLastConstraint c = new WorkAfterLastConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedTimeSpanInt(c, source.getMaxWorkAfterLast(), false);
            target.constraints.add(c);
        }

        if (source.getMinTotalBreakTime() != null) {
            TotalBreakTimeConstraint c = new TotalBreakTimeConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedTimeSpanInt(c, source.getMinTotalBreakTime(), true);
            target.constraints.add(c);
        }

        if (source.getMaxTotalBreakTime() != null) {
            TotalBreakTimeConstraint c = new TotalBreakTimeConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedTimeSpanInt(c, source.getMaxTotalBreakTime(), false);
            target.constraints.add(c);
        }

        if (source.getMinTotalBreakTimeFraction() != null) {
            BreakTimeFractionConstraint c = new BreakTimeFractionConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedDouble(c, source.getMinTotalBreakTimeFraction(), true);
            target.constraints.add(c);
        }

        if (source.getMaxTotalBreakTimeFraction() != null) {
            BreakTimeFractionConstraint c = new BreakTimeFractionConstraint(context.instance);
            c.weightStrategy = context.weight.convertWeightedDouble(c, source.getMaxTotalBreakTimeFraction(), false);
            target.constraints.add(c);
        }

        BreakConfigurations.BreakConfiguration.BreakSet set = source.getBreakSet();

        if (set.isOrdered() == Boolean.TRUE) {
            target.breakSetOrdered = true;
            target.orderedBreaks = new ArrayList<>();
            set.getBreak().forEach(i -> target.orderedBreaks.add(convertBreakSetEntry(i)));
        } else {
            target.unorderedBreaks = new HashMap<>();
            set.getBreak().forEach(i -> target.unorderedBreaks.put(i.getValue(), convertBreakSetEntry(i)));
        }

    }

    /**
     * Converts an entry for a break set.
     * @param br The break entry.
     * @return The converted break occurrence.
     */
    private BreakConfiguration.BreakOccurrence convertBreakSetEntry(BreakConfigurations.BreakConfiguration.BreakSet.Break br) {

        BreakConfiguration.BreakOccurrence occurrence = new BreakConfiguration.BreakOccurrence();
        occurrence.id = br.getValue();

        if (br.getMinCount() != null) {
            occurrence.min = br.getMinCount().intValue();
        }

        if (br.getMaxCount() != null) {
            occurrence.max = br.getMaxCount().intValue();
        }

        return occurrence;

    }

}
