package at.ac.tuwien.dbai.ges.solver.converter.breaks;

import at.ac.tuwien.dbai.ges.schema.BreakTypes;
import at.ac.tuwien.dbai.ges.schema.WeightedTimePoint;
import at.ac.tuwien.dbai.ges.schema.WeightedTimeSpan;
import at.ac.tuwien.dbai.ges.solver.constraint.breaks.*;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionContext;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionException;
import at.ac.tuwien.dbai.ges.solver.converter.Converter;
import at.ac.tuwien.dbai.ges.solver.definition.BreakType;

import javax.xml.bind.JAXBElement;

/**
 * Converts definitions of break types.
 *
 * @author Lucas Kletzander
 */
public class BreakTypeConverter extends Converter<BreakTypes.BreakType, BreakType> {

    public BreakTypeConverter(ConversionContext context) {
        super(BreakTypes.BreakType.class, BreakType.class, context);
    }

    @Override
    public void convert(BreakTypes.BreakType source, BreakType target) throws ConversionException {

        target.id = source.getID();
        BreakLengthConstraint lengthConstraint = new BreakLengthConstraint(context.instance);
        BreakStartConstraint startConstraint = new BreakStartConstraint(context.instance);
        BreakEndConstraint endConstraint = new BreakEndConstraint(context.instance);
        StartShiftConstraint startShift = new StartShiftConstraint(context.instance, source.getID());
        EndShiftConstraint endShift = new EndShiftConstraint(context.instance, source.getID());
        WorkBeforeConstraint beforeConstraint = new WorkBeforeConstraint(context.instance, source.getID());
        WorkAfterConstraint afterConstraint = new WorkAfterConstraint(context.instance, source.getID());

        for (JAXBElement<?> element : source.getMinLengthOrMaxLengthOrMinStart()) {

            switch (element.getName().getLocalPart().toLowerCase()) {
                case "minlength":
                    lengthConstraint.weightStrategy.add(context.weight.convertWeightedTimeSpanInt(lengthConstraint, (WeightedTimeSpan) element.getValue(), true));
                    break;
                case "maxlength":
                    lengthConstraint.weightStrategy.add(context.weight.convertWeightedTimeSpanInt(lengthConstraint, (WeightedTimeSpan) element.getValue(), false));
                    break;
                case "minstart":
                    startConstraint.weightStrategy.add(context.weight.convertWeightedTimePointInt(startConstraint, (WeightedTimePoint) element.getValue(), true));
                    break;
                case "maxstart":
                    startConstraint.weightStrategy.add(context.weight.convertWeightedTimePointInt(startConstraint, (WeightedTimePoint) element.getValue(), false));
                    break;
                case "minend":
                    endConstraint.weightStrategy.add(context.weight.convertWeightedTimePointInt(endConstraint, (WeightedTimePoint) element.getValue(), true));
                    break;
                case "maxend":
                    endConstraint.weightStrategy.add(context.weight.convertWeightedTimePointInt(endConstraint, (WeightedTimePoint) element.getValue(), false));
                    break;
                case "minstartshift":
                    startShift.weightStrategy.add(context.weight.convertWeightedTimeSpanInt(startShift, (WeightedTimeSpan) element.getValue(), true));
                    break;
                case "maxstartshift":
                    startShift.weightStrategy.add(context.weight.convertWeightedTimeSpanInt(startShift, (WeightedTimeSpan) element.getValue(), false));
                    break;
                case "minendshift":
                    endShift.weightStrategy.add(context.weight.convertWeightedTimeSpanInt(endShift, (WeightedTimeSpan) element.getValue(), true));
                    break;
                case "maxendshift":
                    endShift.weightStrategy.add(context.weight.convertWeightedTimeSpanInt(endShift, (WeightedTimeSpan) element.getValue(), false));
                    break;
                //TODO workLength
                case "minworkbefore":
                    beforeConstraint.weightStrategy.add(context.weight.convertHardTimeSpanInt(beforeConstraint, (String) element.getValue(), true));
                    break;
                case "maxworkbefore":
                    beforeConstraint.weightStrategy.add(context.weight.convertHardTimeSpanInt(beforeConstraint, (String) element.getValue(), false));
                    break;
                case "minworkafter":
                    afterConstraint.weightStrategy.add(context.weight.convertHardTimeSpanInt(afterConstraint, (String) element.getValue(), true));
                    break;
                case "maxworkafter":
                    afterConstraint.weightStrategy.add(context.weight.convertHardTimeSpanInt(afterConstraint, (String) element.getValue(), true));
                    break;
                default:
                    throw new ConversionException("Unknown constraint " + element.getName());
            }

        }

        if (!lengthConstraint.weightStrategy.isEmpty()) {
            target.localConstraints.add(lengthConstraint);
        }
        if (!startConstraint.weightStrategy.isEmpty()) {
            target.localConstraints.add(startConstraint);
        }
        if (!endConstraint.weightStrategy.isEmpty()) {
            target.localConstraints.add(endConstraint);
        }
        if (!startShift.weightStrategy.isEmpty()) {
            target.constraints.add(startShift);
        }
        if (!endShift.weightStrategy.isEmpty()) {
            target.constraints.add(endShift);
        }
        if (!beforeConstraint.weightStrategy.isEmpty()) {
            target.constraints.add(beforeConstraint);
        }
        if (!afterConstraint.weightStrategy.isEmpty()) {
            target.constraints.add(afterConstraint);
        }

    }

}
