package at.ac.tuwien.dbai.ges.solver.converter.demand;

import at.ac.tuwien.dbai.ges.schema.TaskCover;
import at.ac.tuwien.dbai.ges.schema.WeightedIntSkill;
import at.ac.tuwien.dbai.ges.schema.WeightedTimePoint;
import at.ac.tuwien.dbai.ges.solver.Util;
import at.ac.tuwien.dbai.ges.solver.constraint.task.TaskEndConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.task.TaskRequirementConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.task.TaskStartConstraint;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionContext;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionException;
import at.ac.tuwien.dbai.ges.solver.converter.Converter;
import at.ac.tuwien.dbai.ges.solver.converter.DateTimeConverter;
import at.ac.tuwien.dbai.ges.solver.demand.TaskDemand;
import at.ac.tuwien.dbai.ges.solver.demand.TaskDemands;
import at.ac.tuwien.dbai.ges.solver.solution.Task;
import at.ac.tuwien.dbai.ges.solver.solution.TaskInstance;

import javax.xml.bind.JAXBElement;
import java.util.ArrayList;
import java.util.stream.IntStream;

/**
 * Converts one task cover and adds the copies of the task request to all specified days in the demand structure.
 *
 * @author Lucas Kletzander
 */
public class TaskDemandConverter extends Converter<TaskCover, TaskDemands> {

    private static int generatedId = 0;

    public TaskDemandConverter(ConversionContext context) {
        super(TaskCover.class, TaskDemands.class, context);
    }

    @Override
    public void convert(TaskCover source, TaskDemands target) throws ConversionException {

        TaskDemand demand = new TaskDemand();
        if (source.getID() == null) {
            demand.id = String.valueOf(generatedId++);
        } else {
            demand.id = source.getID();
        }
        demand.task = source.getTask();
        demand.length = DateTimeConverter.convertTimeSpan(source.getLength());

        demand.constraints = new ArrayList<>();
        TaskStartConstraint startConstraint = new TaskStartConstraint(context.instance);
        TaskEndConstraint endConstraint = new TaskEndConstraint(context.instance);

        for (JAXBElement<WeightedTimePoint> timePoint : source.getStartOrEnd()) {

            switch (timePoint.getName().getLocalPart().toLowerCase()) {
                case "start":
                    startConstraint.weightStrategy.add(context.weight.convertWeightedTimePointInt(startConstraint, timePoint.getValue(), true));//TODO use flag whether to include this constraint or not (to prevent multiple copies of constraint that can't be violated)
                    demand.earliestStart = Util.updateMin(demand.earliestStart, timePoint.getValue());
                    break;
                case "end":
                    endConstraint.weightStrategy.add(context.weight.convertWeightedTimePointInt(endConstraint, timePoint.getValue(), false));
                    demand.latestEnd = Util.updateMax(demand.latestEnd, timePoint.getValue());
                    break;
                default:
                    throw new ConversionException("Unknown constraint " + timePoint.getName());
            }

        }

        if (!startConstraint.weightStrategy.isEmpty()) {
            demand.constraints.add(startConstraint);
        }

        if (!endConstraint.weightStrategy.isEmpty()) {
            demand.constraints.add(endConstraint);
        }

        if (source.getAllowSplit() != null) {
            demand.allowSplit = source.getAllowSplit().isBreaksOnly() == Boolean.TRUE ? TaskDemand.AllowSplit.BREAKS_ONLY : TaskDemand.AllowSplit.YES;
            if (source.getAllowSplit().getMinLength() != null) {
                demand.minSplitLength = DateTimeConverter.convertTimeSpan(source.getAllowSplit().getMinLength());
            }
        }

        if (source.getPrerequisites() != null) {
            demand.prerequisites.addAll(source.getPrerequisites().getTask());
        }

        TaskRequirementConstraint requirementConstraint = new TaskRequirementConstraint(context.instance);

        for (JAXBElement<WeightedIntSkill> req : source.getRequirements().getMinOrMax()) {

            TaskRequirementConstraint constraint = req.getValue().getSkill() == null ? requirementConstraint : demand.skillConstraints.computeIfAbsent(req.getValue().getSkill(), i -> new TaskRequirementConstraint(context.instance));
            switch (req.getName().getLocalPart().toLowerCase()) {
                case "min":
                    constraint.weightStrategy.add(context.weight.convertWeightedInteger(constraint, req.getValue(), true));
                    break;
                case "max":
                    constraint.weightStrategy.add(context.weight.convertWeightedInteger(constraint, req.getValue(), false));
                    break;
                default:
                    throw new ConversionException("Unknown constraint " + req.getName());
            }

        }

        if (!requirementConstraint.weightStrategy.isEmpty()) {
            demand.constraints.add(requirementConstraint);
        }

        boolean[] days = context.dateTime.convertDayOptions(source.getDay(), source.getWeekDay(), source.getDays());
        IntStream.range(0, days.length).filter(i -> days[i]).forEach(i -> {

            TaskDemand d = demand.clone();
            d.id = i + ":" + demand.id;
            target.demand.get(i).add(d);
            target.idMap.put(d.id, d);

            TaskInstance inst = new TaskInstance(d.id, i);
            Task task = new Task();
            task.demandId = d.id;
            task.start = d.earliestStart;
            task.end = d.earliestStart + d.length;
            inst.parts.add(task);
            context.solution.addTaskInstance(inst);

        });

    }

}
