package at.ac.tuwien.dbai.ges.solver.converter.employee;

import at.ac.tuwien.dbai.ges.schema.*;
import at.ac.tuwien.dbai.ges.solver.constraint.employee.VariableEmployeeConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.ShiftArrangementConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.ShiftPairConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.solver.OverlapConstraint;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionContext;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionException;
import at.ac.tuwien.dbai.ges.solver.converter.Converter;
import at.ac.tuwien.dbai.ges.solver.converter.DateTimeConverter;
import at.ac.tuwien.dbai.ges.solver.definition.*;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;

import javax.xml.bind.JAXBElement;
import java.util.LinkedList;

/**
 * Converts employee definitions and corresponding constraints.
 *
 * @author Lucas Kletzander
 */
public class EmployeeConverter extends Converter<Employees, EmployeeDefinition> {

    public EmployeeConverter(ConversionContext context) {
        super(Employees.class, EmployeeDefinition.class, context);
    }

    @Override
    public void convert(Employees source, EmployeeDefinition target) throws ConversionException {

        for (AbstractEmployeeType type : source.getEmployeeList().getEmployeeOrVariableEmployee()) {

            AbstractEmployee employee;

            if (type.getClass().equals(EmployeeList.Employee.class)) {

                Employee e = new Employee();
                e.id = ((EmployeeList.Employee) type).getID();
                e.name = ((EmployeeList.Employee) type).getName();
                target.employees.put(e.id, e);
                employee = e;

                OverlapConstraint overlap = new OverlapConstraint(context.instance);
                overlap.strategy = context.weight.convertHardInt(overlap);
                EmployeeSchedule schedule = new EmployeeSchedule(e.id, context.instance.history + context.instance.periodLength, overlap);
                context.solution.addEmployeeSchedule(schedule);//TODO usage optional

            } else {

                EmployeeList.VariableEmployee var = (EmployeeList.VariableEmployee) type;
                VariableEmployee e = new VariableEmployee();

                e.id = var.getID();
                if (var.getCost() != null) {
                    e.cost = var.getCost();
                }
                target.varEmployees.put(e.id, e);
                target.employeeConstraints.add(context.provider.getConverter(EmployeeList.VariableEmployee.class, VariableEmployeeConstraint.class).createAndConvert(type));
                employee = e;

            }

            if (type.getContracts() != null) {

                employee.contracts.addAll(type.getContracts().getContract());

                LinkedList<String> newContracts = new LinkedList<>(type.getContracts().getContract());

                while (!newContracts.isEmpty()) {
                    Contract contract = context.instance.contractDefinition.contracts.get(newContracts.poll());
                    if (contract.parent != null && !employee.contracts.contains(contract.parent)) {
                        employee.contracts.add(contract.parent);
                        newContracts.add(contract.parent);
                    }
                }

            }

            if (type.getPreferences() != null) {
                for (Object preference : type.getPreferences().getPenalizedTaskOrShiftOffRequestOrShiftOnRequest()) {
                    employee.preferences.add(context.provider.getConverter(preference.getClass(), ShiftArrangementConstraint.class).createAndConvert(preference));
                    //TODO penalized task missing
                }
            }

            if (type.getPreAssignments() != null) {

                for (Object assignment : type.getPreAssignments().getShiftOrNoShift()) {

                    if (assignment instanceof PreAssignments.Shift) {

                        PreAssignments.Shift s = (PreAssignments.Shift) assignment;

                        if (s.getExtend() == null || s.getExtend().toLowerCase().equals("none")) {

                            Shift shift = new Shift();
                            shift.type = s.getID();
                            if (s.getStartTime() != null) {//TODO start value if not fixed / extend?
                                shift.start = DateTimeConverter.convertTimePoint(s.getStartTime());
                                shift.startFixed = true;
                            }
                            if (s.getEndTime() != null) {
                                shift.end = DateTimeConverter.convertTimePoint(s.getEndTime());
                                shift.endFixed = true;
                            }
                            //TODO fixed tasks / skills / breaks
                            int day = context.dateTime.convertDateOrDay(s.getDay());
                            context.solution.schedule.get(employee.id).shifts[day] = shift;//TODO does not work for abstract employee
                            context.solution.schedule.get(employee.id).fixed[day] = true;

                        } else {
                            //TODO
                        }

                    } else {
                        employee.preferences.add(context.provider.getConverter(assignment.getClass(), ShiftArrangementConstraint.class).createAndConvert(assignment));
                    }

                }

            }

        }

        if (source.getEmployeePairings() != null) {
            for (JAXBElement<EmployeePair> pair : source.getEmployeePairings().getPairOrNotPair()) {
                target.pairConstraints.add(context.provider.getConverter(pair, ShiftPairConstraint.class).createAndConvert(pair.getValue()));
            }
        }

    }

}
