package at.ac.tuwien.dbai.ges.solver.converter.shift;

import at.ac.tuwien.dbai.ges.schema.*;
import at.ac.tuwien.dbai.ges.solver.Util;
import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.ShiftConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.EmployeeLimitConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.ShiftEndConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.ShiftLengthConstraint;
import at.ac.tuwien.dbai.ges.solver.constraint.shift.ShiftStartConstraint;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionContext;
import at.ac.tuwien.dbai.ges.solver.converter.ConversionException;
import at.ac.tuwien.dbai.ges.solver.converter.Converter;
import at.ac.tuwien.dbai.ges.solver.converter.DateTimeConverter;
import at.ac.tuwien.dbai.ges.solver.definition.ShiftDefinition;
import at.ac.tuwien.dbai.ges.solver.definition.ShiftType;

import javax.xml.bind.JAXBElement;
import java.util.ArrayList;
import java.util.List;

/**
 * Converts shift definitions and corresponding constraints.
 *
 * @author Lucas Kletzander
 */
public class ShiftConverter extends Converter<Shifts, ShiftDefinition> {

    public ShiftConverter(ConversionContext context) {
        super(Shifts.class, ShiftDefinition.class, context);
    }

    @Override
    public void convert(Shifts source, ShiftDefinition target) throws ConversionException {

        for (ShiftTypes.ShiftType s : source.getShiftTypes().getShiftType()) {

            ShiftType shiftType = new ShiftType();
            shiftType.id = s.getID();
            shiftType.name = s.getName();
            shiftType.label = s.getLabel();

            int minLength = Integer.MAX_VALUE;
            int maxLength = -1;
            WeightedTimeSpan span;

            for (JAXBElement<?> element : s.getMinStartTimeOrMaxStartTimeOrMinEndTime()) {

                switch (element.getName().getLocalPart().toLowerCase()) {
                    case "minstarttime":
                        shiftType.shiftTypeConstraints.add(convertStartTime((WeightedTimePoint) element.getValue(), true));
                        shiftType.earliestStart = Util.updateMin(shiftType.earliestStart, (WeightedTimePoint) element.getValue());
                        break;
                    case "maxstarttime":
                        shiftType.shiftTypeConstraints.add(convertStartTime((WeightedTimePoint) element.getValue(), false));
                        shiftType.latestStart = Util.updateMax(shiftType.latestStart, (WeightedTimePoint) element.getValue());
                        break;
                    case "minendtime":
                        shiftType.shiftTypeConstraints.add(convertEndTime((WeightedTimePoint) element.getValue(), true));
                        shiftType.earliestEnd = Util.updateMin(shiftType.earliestEnd, (WeightedTimePoint) element.getValue());
                        break;
                    case "maxendtime":
                        shiftType.shiftTypeConstraints.add(convertEndTime((WeightedTimePoint) element.getValue(), false));
                        shiftType.latestEnd = Util.updateMax(shiftType.latestEnd, (WeightedTimePoint) element.getValue());
                        break;
                    case "minlength":
                        span = (WeightedTimeSpan) element.getValue();
                        shiftType.shiftTypeConstraints.add(convertLength(span, true));
                        if (span.getWeight() == null) {
                            minLength = Math.min(DateTimeConverter.convertTimeSpan(span.getValue()), minLength);
                        }
                        break;
                    case "maxlength":
                        span = (WeightedTimeSpan) element.getValue();
                        shiftType.shiftTypeConstraints.add(convertLength(span, false));
                        if (span.getWeight() == null) {
                            maxLength = Math.max(DateTimeConverter.convertTimeSpan(span.getValue()), maxLength);
                        }
                        break;
                    default:
                        throw new ConversionException("Unknown shift constraint " + element.getName());
                }

            }

            if (minLength < Integer.MAX_VALUE) {
                shiftType.earliestEnd = Math.max(shiftType.earliestEnd, shiftType.earliestStart + minLength);
            }
            if (maxLength > -1) {
                shiftType.latestEnd = shiftType.latestEnd == -1 ? shiftType.latestStart + maxLength : Math.min(shiftType.latestEnd, shiftType.latestStart + maxLength);
            }

            if (s.getEmployeeLimits() != null) {

                Converter<ShiftTypes.ShiftType.EmployeeLimits.EmployeeLimit, EmployeeLimitConstraint> converter = context.provider.getConverter(ShiftTypes.ShiftType.EmployeeLimits.EmployeeLimit.class, EmployeeLimitConstraint.class);

                for (ShiftTypes.ShiftType.EmployeeLimits.EmployeeLimit limit : s.getEmployeeLimits().getEmployeeLimit()) {
                    shiftType.shiftTypeConstraints.add(converter.createAndConvert(limit));
                }

            }

            shiftType.validDays = context.dateTime.convertDayCollection(s.getValidDays());

            //TODO convert fixed tasks

            target.shiftTypes.put(shiftType.id, shiftType);
            List<String> singleGroup = new ArrayList<>();
            singleGroup.add(shiftType.id);
            target.shiftGroups.put(shiftType.id, singleGroup);

        }

        if (source.getShiftGroups() != null) {

            for (ShiftGroups.ShiftGroup g : source.getShiftGroups().getShiftGroup()) {

                List<String> group = new ArrayList<>(g.getShift());
                target.shiftGroups.put(g.getID(), group);

            }

        }

        if (source.getGlobalShiftConstraints() != null) {

            for (Object c : source.getGlobalShiftConstraints().getAverageShiftLengthOrShiftInstances()) {
                target.globalShiftConstraints.add(context.provider.getConverter(c.getClass(), ShiftConstraint.class).createAndConvert(c));
            }

        }

        if (source.isAllowBorderShifts() != null) {
            target.allowBorderShifts = source.isAllowBorderShifts();
        }

        if (source.getTimeSlotLength() != null) {
            context.instance.shiftSlotLength = DateTimeConverter.convertTimeSpan(source.getTimeSlotLength());
        }

    }

    /**
     * Converts a shift start constraint.
     * @param startTime The limit for the start time.
     * @param isMin Whether the limit is a minimum.
     * @return The shift start constraint.
     */
    public ShiftStartConstraint convertStartTime(WeightedTimePoint startTime, boolean isMin) {

        ShiftStartConstraint constraint = new ShiftStartConstraint(context.instance);
        constraint.strategy = context.weight.convertWeightedTimePointInt(constraint, startTime, isMin);
        return constraint;

    }

    /**
     * Converts a shift end constraint.
     * @param endTime The limit for the end time.
     * @param isMin Whether the limit is a minimum.
     * @return The shift end constraint.
     */
    public ShiftEndConstraint convertEndTime(WeightedTimePoint endTime, boolean isMin) {

        ShiftEndConstraint constraint = new ShiftEndConstraint(context.instance);
        constraint.strategy = context.weight.convertWeightedTimePointInt(constraint, endTime, isMin);
        return constraint;

    }

    /**
     * Converts a shift length constraint.
     * @param length The limit for the length.
     * @param isMin Whether the limit is a minimum.
     * @return The shift length constraint.
     */
    public ShiftLengthConstraint convertLength(WeightedTimeSpan length, boolean isMin) {

        ShiftLengthConstraint constraint = new ShiftLengthConstraint(context.instance);
        constraint.strategy = context.weight.convertWeightedTimeSpanInt(constraint, length, isMin);
        return constraint;

    }

}
