package at.ac.tuwien.dbai.ges.solver.definition;

import at.ac.tuwien.dbai.ges.solver.constraint.hierarchy.*;
import at.ac.tuwien.dbai.ges.solver.demand.ShiftDemands;
import at.ac.tuwien.dbai.ges.solver.demand.TaskDemands;
import at.ac.tuwien.dbai.ges.solver.solution.Break;
import at.ac.tuwien.dbai.ges.solver.solution.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.solver.solution.Shift;
import at.ac.tuwien.dbai.ges.solver.solution.Solution;

import java.util.Set;
import java.util.stream.Stream;

/**
 * A complete problem instance including methods to access various types of constraints.
 *
 * @author Lucas Kletzander
 */
public class Instance {

    public String instanceId;
    public int periodLength;
    public int history;
    public int timeSlotLength = 1;
    public int shiftSlotLength = -1;
    public int firstWeekDay = 0;
    public int startTime = 0;
    public int endTime = 24 * 60;
    public Weekend weekend = new Weekend();
    public boolean ignoreIndividualRosters = false;
    public boolean shiftDesign = false;
    public CyclicSchedule cyclicSchedule = CyclicSchedule.NONE;
    public boolean cutoffStart = false;//TODO is this used in all relevant constraints?
    public boolean cutoffEnd = false;

    public TaskDefinition taskDefinition = new TaskDefinition();
    public ShiftDefinition shiftDefinition = new ShiftDefinition();
    public BreakDefinition breakDefinition = new BreakDefinition();
    public SkillDefinition skillDefinition = new SkillDefinition();
    public ContractDefinition contractDefinition = new ContractDefinition();
    public EmployeeDefinition employeeDefinition = new EmployeeDefinition();
    public DemandType demandType;
    public ShiftDemands shiftDemands;
    public Void timeDemands;//TODO not yet implemented
    public TaskDemands taskDemands;

    public enum CyclicSchedule {
        NONE, CYCLIC, ROTATING
    }

    public class Weekend {

        public boolean isSimple = true;
        public boolean[] isWeekend = new boolean[] {false, false, false, false, false, true, true};
        public int weekendStartSimple;
        public int startWeekDay;
        public int startTime;
        public boolean startCompleteOnly;
        public int endWeekDay;
        public int endTime;
        public boolean endCompleteOnly;

        public void updateWeekendStartSimple() {

            boolean foundTrue = false;

            for (int i = 6; i >= 0; i--) {

                if (isWeekend[i]) {
                    foundTrue = true;
                } else if (foundTrue) {
                    weekendStartSimple = i + 1;
                }

                weekendStartSimple = 0;

            }

        }

        public int isWeekend(Shift shift, int day) {

            int weekDay = getWeekDay(day);
            if (isSimple) {
                if (isWeekend[weekDay]) {
                    return (day + firstWeekDay - weekendStartSimple) / 7;
                } else {
                    return -1;
                }
            }

            int start = startCompleteOnly ? shift.start : shift.end;
            int end = endCompleteOnly ? shift.end : shift.start;

            boolean isWeekend;
            if (startWeekDay <= endWeekDay) {
                isWeekend = compare(weekDay, start, startWeekDay, startTime) >= 0 && compare(weekDay, end, endWeekDay, endTime) <= 0;
            } else {
                weekDay += weekDay <= endWeekDay ? 7 : 0;
                isWeekend = compare(weekDay, start, startWeekDay, startTime) >= 0 && compare(weekDay, end, endWeekDay + 7, endTime) <= 0;
            }

            if (isWeekend) {
                return (day + start / (24 * 60) + firstWeekDay - startWeekDay) / 7;
            } else {
                return -1;
            }

        }

    }

    public int getWeekDay(int day) {
        return (day + firstWeekDay) % 7;
    }

    public int compare(int day1, int time1, int day2, int time2) {
        return time1 - (day2 - day1) * 24 * 60 - time2;
    }

    public int timeToTimeSlots(int time, boolean roundUp) {
        return (time + (roundUp ? timeSlotLength - 1 : 0)) / timeSlotLength;
    }

    public int getMaxDay() {
        return history + periodLength;
    }

    public int getMaxTime() {
        return (history + periodLength) * 24 * 60;
    }

    public Stream<BreakConstraint> getBreakConstraintStream(Break br) {
        return breakDefinition.getBreakConstraints(br);
    }

    public Stream<BreakArrangementConstraint> getBreakArrangementConstraintStream(EmployeeSchedule schedule, int day) {
       return breakDefinition.getBreakArrangementConstraints(schedule.id, day, schedule.shifts[day]);
    }

    public Stream<EmployeeConstraint> getEmployeeConstraintStream() {
        return employeeDefinition.getEmployeeConstraints();
    }

    public Stream<ShiftConstraint> getShiftConstraintStream(String type, AbstractEmployee employee, int day) {//TODO split into shift shape and shift assignment constraints for performance?
        Stream<ShiftConstraint> stream = Stream.concat(shiftDefinition.getShiftConstraints(type), contractDefinition.getShiftConstraints(employee.contracts));
        return demandType == DemandType.SHIFT ? Stream.concat(stream, shiftDemands.getShiftConstraints(type, day, employee.skills)) : stream;
    }

    public Stream<ShiftArrangementConstraint> getShiftArrangementConstraintStream(AbstractEmployee employee) {
        return Stream.concat(contractDefinition.getShiftArrangementConstraints(employee.contracts), employeeDefinition.getShiftArrangementConstraints(employee.id));
    }

    public Stream<TaskConstraint> getTaskConstraintStream(String id, Set<String> skills) {
        return demandType == DemandType.TASK ? taskDemands.getTaskConstraints(id, skills) : Stream.empty();
    }

    public Stream<Constraint> getConstraintStream() {

        Stream<Constraint> stream = Stream.concat(
                Stream.concat(
                        Stream.concat(breakDefinition.getConstraints(), employeeDefinition.getConstraints()),
                        shiftDefinition.getConstraints()),
                contractDefinition.getConstraints());

        switch (demandType) {
            case SHIFT:
                return Stream.concat(stream, shiftDemands.getConstraints());
            case TIME:
                return stream;//TODO time demands
            case TASK:
                return Stream.concat(stream, taskDemands.getConstraints());
        }

        return stream;

    }

    public Stream<Constraint> getConstraintStream(Solution solution) {
        return Stream.concat(getConstraintStream(), solution.getConstraintStream());
    }

    public enum DemandType {
        SHIFT, TIME, TASK
    }

}
